from __future__ import print_function

import numpy as np
import scipy

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('find_dof_map3_h1')

from petram.helper.matrix_file import write_matrix, write_vector

from petram.mfem_config import use_parallel
if use_parallel:
   from mpi4py import MPI
   num_proc = MPI.COMM_WORLD.size
   myid     = MPI.COMM_WORLD.rank
   from petram.helper.mpi_recipes import *
   import mfem.par as mfem
else:
   import mfem.ser as mfem
   num_proc = 1
   myid = 0
   def allgather(x): return [x]

mapper_debug = False
debug_gf = False

def find_dof_map_h1(idx1, idx2, transu, transv, transu2, transv2, 
                    fes, engine, tdof, tol=1e-4):

   '''
   find_dof_map (H1 element)

   projection coupling between DoFs. so far, it is tested
   only for triangle Nedelec boundary elemets.

   it maps element nodes of on bdr1 (attribute = idx1) to 
   bdr2(attribute =idx2)

   this version accept two mapping function which mappps
   (x, y, z) to (u, v) plane

   Need to test that this approach works even when two 
   surface is not parallel...

   In this version, returned matrix is interpolation constraints P.
   P will be used as follows.
    P^t A P y = P^t f, x = Py
       
   For example..
      if  x2 = x1 and   [A B][x3]  = [x5]
                        [C D][x4]    [x6]
      P should be filled as follows. It starts from identity 
      sparse  matrix and changes the following matrix elements.

      [ 1            ][y1]    [x1]
      [ 1 0          ][y2]    [x2]
      [     1        ][y3]  = [x3] 
      [        1     ][y4]    [x4]
      [     A  B 0   ][y5]    [x5]
      [     C  D   0 ][y6]    [x6]

    row and col of  P^t A P corresponding to x2, x5, x6 will
    be all zero.

    This module actually assemble the transpose of P.
    Diagonal elements should be handled in caller. See em3d_floquet

   '''
   if debug_gf:
      gf = engine.new_gf(fes = fes)
      dprint1(gf.Size())
   
#   if trans is None:
#       trans = np.array([[0, 1, 0], [0, 0, 1]]) # projection to y, z         
   def get_shape_all(fes, k1):
       tr1 = fes.GetBdrElementTransformation(k1)
       el = fes.GetBE(k1)
       nodes1 = el.GetNodes()
       v = mfem.Vector(nodes1.GetNPoints())
       shape = []
       for idx in range(nodes1.GetNPoints()):
          el.CalcShape(nodes1.IntPoint(idx), v)
          shape.append(v.GetDataArray()[idx])
       shape = np.stack(shape)    # shape seems all 1, length = num int-point
       return shape

   def get_all_intpoint2(fes, pt1all, pt2all, pto1all, pto2all, k1all,
                         k2all, sh1all, sh2all, map_1_2,
                         tru, trv, tru2, trv2, 
                         tol, tdof, fesize):

       pt = []
       subvdofs1 = []
       subvdofs2 = []

       map = scipy.sparse.lil_matrix((fesize, fesize), dtype=float)
       num_entry = 0
       num_pts = 0

       decimals = int(np.abs(np.log10(tol)))       
       for k0 in range(len(pt1all)):
           k2 = map_1_2[k0]
           pt1 = pt1all[k0]
           pto1 = pto1all[k0]
           newk1 = k1all[k0]
           sh1 = sh1all[k0]           
           pt2 = pt2all[k2]
           pto2 = pto2all[k2]
           newk2 = k2all[k2]
           sh2 = sh2all[k2]
           #if myid == 1: print newk1
           for k, p in enumerate(pt1):
               num_pts = num_pts + 1
               if newk1[k,2] == -1: continue # not owned by the node
               if newk1[k,2] in tdof: continue
               if newk1[k,2] in subvdofs1: continue               
               dist = np.sum((pt2-p)**2, 1)
               d = np.where(dist == np.min(dist))[0]
               #dprint1('min_dist', np.min(dist))
               if len(d) == 1:
                   d = d[0]
                   s1 = sh1[newk1[k, 0]]
                   s2 = sh2[newk2[d, 0]]
                   #dprint1("case1 ", s1, s2) this looks all 1
                   if s1/s2 < 0: dprint1("not positive")

                   map[newk2[d][2], newk1[k][2]] = np.around(s1/s2, decimals)
                   if debug_gf:
                      gf[newk1[k][1]]= p[0]
                      gf[newk2[d][1]]= -p[0]                      
                   num_entry = num_entry + 1
               else:
                   raise AssertionError("more than two dofs at same plase is not asupported. ")
                   pass
           subvdofs1.extend([s for k, v, s in newk1])
           subvdofs2.extend([s for k, v, s in newk2])
       #print len(subvdofs1), len(subvdofs2)

       total_entry = sum(allgather(num_entry))
       total_pts = sum(allgather(num_pts))
       
       dprint1("map size", map.shape)       
       dprint1("local pts/entry", num_pts, " " , num_entry)
       dprint1("total pts/entry", total_pts, " " , total_entry)       
       return map


   def find_el_center(fes, ibdr1, tru, trv):
       if len(ibdr1) == 0: return np.empty(shape=(0,2))
       mesh = fes.GetMesh()
       pts = np.vstack([np.mean([(tru(mesh.GetVertexArray(kk)),
                                  trv(mesh.GetVertexArray(kk)))
                                  for kk in mesh.GetBdrElementVertices(k)],0)
                        for k in ibdr1])
       return pts

   def get_shape_all2(fes, ibdr1):
       if len(ibdr1) == 0:
           tmp = np.stack([get_shape_all(fes, k) for k in [0]])
           res = np.empty(shape=(0, tmp.shape[1]))
       else:
           res =  np.stack([get_shape_all(fes, k) for k in ibdr1])
       dprint1('get_shape_all2', res.shape)
       return res
        
   def get_element_data(k1, debug, tru, trv):
       tr1 = fes.GetBdrElementTransformation(k1)
       nodes1 = fes.GetBE(k1).GetNodes()
       vdof1 = fes.GetBdrElementVDofs(k1)

       pt1 = np.vstack([np.array([tru(tr1.Transform(nodes1.IntPoint(kk))),
                                  trv(tr1.Transform(nodes1.IntPoint(kk)))])
              for kk in range(len(vdof1))])
       pt1o = np.vstack([np.array(tr1.Transform(nodes1.IntPoint(kk)))
                       for kk in range(len(vdof1))])

       subvdof1 = [x if x>= 0 else -1-x for x in vdof1]

       if use_parallel:
           subvdof2= [fes.GetLocalTDofNumber(i)
                      for i in subvdof1]


           flag = False
           for k, x in enumerate(subvdof2):
               if x >=0:
                  subvdof2[k] = fes.GetMyTDofOffset()+ x
               else:
                  if debug:
                     print('not own')
                     flag = True
           if flag: dprint1(subvdof1, vdof1, subvdof2)

           ## note subdof2 = -1 if it is not owned by the node
       else:
           subvdof2 = subvdof1
       newk1 = np.vstack([(k, xx[0], xx[1])
                for k, xx in enumerate(zip(vdof1, subvdof2))])
       pt1 =  np.vstack([pt1[kk] for kk, v, s in newk1])
       pt1o = np.vstack([pt1o[kk] for kk, v, s in newk1])

       return pt1, newk1, pt1o

   def resolve_nonowned_dof(pt1all, pt2all, k1all, k2all, map_1_2):
       '''
       resolves shadowed DoF
       this is done based on integration point distance.
       It searches a closeest true (non-shadow) DoF point
       '''
       for k in range(len(pt1all)):
          subvdof1 = k1all[k][:,2]
          k2 = map_1_2[k]
          subvdof2 = k2all[k2][:,2]
          pt2 = pt2all[k2]
          check = False
          if -1 in subvdof2:
                check = True
                dprint1('before resolving dof', subvdof2)
          for kk, x in enumerate(subvdof2):
             if x == -1:
                dist = pt2all-pt2[kk]
                dist = np.sqrt(np.sum((dist)**2, -1))
                fdist= dist.flatten()
                isort = np.argsort(fdist)
                minidx =  np.where(dist.flatten() == np.min(dist.flatten()))[0]
                #dprint1("distances", np.min(dist),fdist[isort[:25]])
                while all(k2all[:,:,2].flatten()[minidx] == -1):
                    dprint1("distances (non -1 exists?)", fdist[minidx],
                            k2all[:,:,2].flatten()[minidx])
                    minidx = np.hstack((minidx, isort[len(minidx)]))
                dprint1("distances", np.min(dist),fdist[minidx],
                        fdist[isort[:len(minidx)+1]])                    
                dprint1("minidx",  minidx, k2all[:,:,2].flatten()[minidx])
                for i in minidx:
                    if k2all[:,:,2].flatten()[i] != -1:
                       subvdof2[kk] = k2all[:,:,2].flatten()[i]

          if check:
              dprint1('resolved dof', k2all[k2][:,2])
              if -1 in subvdof2: raise AssertionError("failed to resolve shadow DoF")

   nbe = fes.GetNBE()
   ibdr1 = [i for i in range(nbe) if fes.GetBdrAttribute(i) in idx1]
   ibdr2 = [i for i in range(nbe) if fes.GetBdrAttribute(i) in idx2]

   ct1 = find_el_center(fes, ibdr1, transu, transv)
   ct2 = find_el_center(fes, ibdr2, transu2, transv2)

   arr1 = [get_element_data(k, False, transu, transv) for k in ibdr1]
   arr2 = [get_element_data(k, False, transu2, transv2) for k in ibdr2]

   
   sh1all = get_shape_all2(fes, ibdr1)
   sh2all = get_shape_all2(fes, ibdr2)

   # prepare default shape...
   ttt = [get_element_data(k, False, transu, transv) for k in [0]]
   ptzero = np.stack([x for x, y, z in ttt])
   ptozero = np.stack([z for x, y, z in ttt])
   kzero = np.stack([y for x, y, z in ttt])
   
   # pt is on (u, v), pto is (x, y, z)
   pt1all = (np.stack([x for x, y, z in arr1]) if len(arr1)!=0 else
             np.empty(shape=(0,ptzero.shape[1],ptzero.shape[2])))
   pt2all = (np.stack([x for x, y, z in arr2]) if len(arr2)!=0 else
             np.empty(shape=(0,ptzero.shape[1],ptzero.shape[2])))            
   pto1all = (np.stack([z for x, y, z in arr1]) if len(arr1)!=0 else
             np.empty(shape=(0,ptozero.shape[1],ptozero.shape[2])))
   pto2all = (np.stack([z for x, y, z in arr2]) if len(arr2)!=0 else
             np.empty(shape=(0,ptozero.shape[1],ptozero.shape[2])))            

   if use_parallel:
      if MPI.INT.size == 4:
          dtype = np.int32
      else:
          dtype = np.int64
   else:
      dtype = np.int32
   k1all = (np.stack([y for x, y, z in arr1]).astype(dtype) if len(arr1)!=0
            else np.empty(shape=(0,kzero.shape[1],kzero.shape[2])).astype(dtype))
   k2all = (np.stack([y for x, y, z in arr2]).astype(dtype) if len(arr2)!=0
            else np.empty(shape=(0,kzero.shape[1],kzero.shape[2])).astype(dtype))

   if mapper_debug:         
   #   for i in range(num_proc):
      for i in range(1):
          MPI.COMM_WORLD.Barrier()      
          if ( myid == 1): 
             dprint3("checking ptall ", myid)
             for k, x in enumerate(pt1all):
                print(x, k1all[k], sh1all[k])
      for i in range(num_proc):
          MPI.COMM_WORLD.Barrier()      
          if ( myid == i): 
             dprint3("checking ptall2 ", myid)
             for k, x in enumerate(pt2all):
                 print(x, k2all[k], sh2all[k])             


   fesize = fes.GetNDofs()
   if use_parallel:
       # share ibr2 (destination information among nodes...)
       ct2 =  allgather_vector(ct2, MPI.DOUBLE)
       pt2all =  allgather_vector(pt2all, MPI.DOUBLE)
       pto2all = allgather_vector(pto2all, MPI.DOUBLE)
       k2all =  allgather_vector(k2all, MPI.INT)
       sh2all =  allgather_vector(sh2all, MPI.DOUBLE)
       fesize = fes.GlobalTrueVSize()

   dprint3('k2all', k2all.shape)
   dprint3('sh2all', sh2all.shape)
   
   ctr_dist = np.array([np.min(np.sum((ct2-c)**2, 1)) for c in ct1])
   if ctr_dist.size > 0 and np.max(ctr_dist) > 1e-15:
       print('Center Dist may be too large (check mesh): ' + str(np.max(ctr_dist)))
       
   map_1_2= [np.argmin(np.sum((ct2-c)**2, 1)) for c in ct1]
   if use_parallel:
       resolve_nonowned_dof(pt1all, pt2all, k1all, k2all, map_1_2)

   map =  get_all_intpoint2(fes, pt1all, pt2all, pto1all, pto2all, k1all,
                            k2all, sh1all, sh2all, map_1_2,
                            transu, transv,
                            transu2, transv2,
                            tol, tdof, fesize)

   if debug_gf:
       write_vector("h1_map", gf)

   if debug.debug_default_level > 1:
      if use_parallel: 
          for i in range(num_proc):
             MPI.COMM_WORLD.Barrier()      
             if ( myid == i ): 
                dprint3("checking map on node ", myid)
                dprint3(map)
      else:
          pass

   return map



'''

 Direct Wrappwer Call provides the access to low level (PyMFEM)
 functionality during a Petra-M simulation

'''
import numpy as np

from petram.mfem_config import use_parallel
if use_parallel:
    from petram.helper.mpi_recipes import *
    import mfem.par as mfem
else:
    import mfem.ser as mfem


class DWC(object):
    def __init__(self):
        pass

    def make_args(self, mode, kwargs):
        if mode == 'postprocess':
            return tuple(), kwargs
        elif mode == 'timestep':
            t = kwargs.pop('time')
            return (t, ), kwargs
        elif mode == 'checkpoint':
            t = kwargs.pop('time')
            icp = kwargs.pop('icheckpoint')
            return (t, icp), kwargs
        elif mode == 'loopcontrol':
            ct = kwargs.pop('count')
            return (ct,), kwargs
        elif mode == 'nlcheckpoint':
            ct = kwargs.pop('count')
            return (ct,), kwargs
        elif mode == 'init':
            return tuple(), kwargs
        else:
            # default no argument
            return tuple(), kwargs

    def call(self, caller, *args, **kwargs):
        '''
        general call
        '''
        raise NotImplementedError("call must be implemented by a user")

    def loopcontrol(self, caller, count, *args, **kwargs):
        raise NotImplementedError("loopcontrol must be implemented by a user")

    def postprocess(self, caller,  *args, **kwargs):
        '''
        postprocess is called from solvestep after store_sol
        '''
        raise NotImplementedError("postprocess must be implemented by a user")

    def timestep(self, caller, t, *args, **kwargs):
        '''
        timestep is called from time-dependent solver at every time step
        t = current time
        '''
        raise NotImplementedError("timestep must be implemented by a user")

    def checkpoint(self, caller, t, cp, *args, **kwargs):
        '''
        timestep is called from time-dependent solver at checkpoint
        t = current time
        cp = check pioint index
        '''
        raise NotImplementedError("checkpoint must be implemented by a user")

    def nl_checkpoint(self, caller, count, *args, **kwargs):
        '''
        called during nonlinear iteration loop
        return value: 
             False: normal
             True: emergency stop of iteration
        '''
        raise NotImplementedError(
            "nl_checkpoint must be implemented by a user")

    def nl_start(self, caller, *args, **kwargs):
        '''
        called at the begining of non-linear iteration
        '''
        raise NotImplementedError("nl_start must be implemented by a user")

    def nl_end(self, caller, *args, **kwargs):
        '''
        called at the end of non-linear iteration
        '''
        raise NotImplementedError("nl_end must be implemented by a user")

    def init(self, caller, *args, **kwargs):
        '''
        init is called from init_setting
        '''
        raise NotImplementedError("init must be implemented by a user")


# sample DWC class (see em3d_TE8.pfz)
class Eval_E_para(DWC):
    def __init__(self, faces):
        DWC.__init__(self)
        self.faces = faces

    def postprocess(self, caller, gf=None, edges=None):

        from petram.helper.mpi_recipes import safe_flatstack
        from mfem.common.mpi_debug import nicePrint
        if edges is None:
            return

        print("postprocess is called")
        gfr, gfi = gf
        print(caller, gfr)
        try:
            fes = gfr.ParFESpace()
            mesh = fes.GetParMesh()
        except:
            fes = gfr.FESpace()
            mesh = fes.GetMesh()
        from petram.mesh.mesh_utils import get_extended_connectivity
        if not hasattr(mesh, 'extended_connectivity'):
            get_extended_connectivity(mesh)
        l2e = mesh.extended_connectivity['line2edge']
        idx = safe_flatstack([l2e[e] for e in edges])
        if len(idx) > 0:
            dofs = safe_flatstack([fes.GetEdgeDofs(i) for i in idx])
            size = dofs.size//idx.size

            w = []
            for i in idx:
                # don't put this Tr outside the loop....
                Tr = mesh.GetEdgeTransformation(i)
                w.extend([Tr.Weight()]*size)
            w = np.array(w)
            data = gfr.GetDataArray()[dofs] + 1j*gfi.GetDataArray()[dofs]
            field = data/w
        else:
            w = np.array([])
            field = np.array([])
        nicePrint(w)
        nicePrint(field)



def run_serial(path='', debug=0, thread=True):
    '''
    debug keyword will overwrite debug level setting 
    in model file
    '''
    import subprocess as sp
    import sys
    import mfem
    import os
    from threading import Thread
    import time

    try:
        from Queue import Queue, Empty
    except ImportError:
        from queue import Queue, Empty  # python 3.x

    ON_POSIX = 'posix' in sys.builtin_module_names

    def enqueue_output(p, queue):
        while True:
            line = p.stdout.readline()
            queue.put(line)
            if p.poll() is not None:
                queue.put(p.stdout.read())
                break
        queue.put('process terminated')

    del_path = False

    if path == '':
        if model.param.eval('sol') is None:
            folder = model.scripts.helpers.make_new_sol()
        else:
            folder = model.param.eval('sol')
            folder.clean_owndir()
        path = os.path.join(folder.owndir(), 'model.pmfm')
        model.scripts.helpers.save_model(path)
        m = model.param.getvar('mfem_model')
        try:
            m.generate_script(dir=folder.owndir())
        except:
            import traceback
            traceback.print_exc()
            return

        del_path = True

    import petram
    #from petram.helper.driver_path import serial as driver
    opath = os.getcwd()

    # This is to test driver locally
    #args = [driver, str(path), str(debug)]
    os.chdir(folder.owndir())
    args = [sys.executable, '-u', 'model.py', '-s', '-d', str(debug)]
    p = sp.Popen(args, stdout=sp.PIPE, stderr=sp.STDOUT)

    line = ''
    thread = True
    if thread:
        q = Queue()
        t = Thread(target=enqueue_output, args=(p, q))
        t.daemon = True  # thread dies with the program
        t.start()
        while(True):
            try:
                line = q.get_nowait()  # or q.get(timeout=.1)
            except Empty:
                time.sleep(1.0)
                pass  # print('no output yet')
            else:
                if isinstance(line, bytes):
                    line = line.decode('utf-8')
                    line = '\n'.join(
                        [x for x in line.split('\n') if len(x) > 0])
                print(line)
            if line == 'process terminated':
                break
    else:
        stdoutdata, stderrdata = p.communicate()
        print(stdoutdata)

    p.kill()

    globals()['default_sol_path'] = os.path.dirname(path)
    globals()['default_glvis_args'] = []

    from petram.sol.solsets import read_sol, find_solfiles
    path = model.param.eval('sol').owndir()
    try:
        solfiles = find_solfiles(path=path)
        model.variables.setvar('solfiles', solfiles)
    except:
        model.variables.delvar('solfiles')

    os.chdir(opath)


ans(run_serial(*args, **kwargs))




'''
   Copyright (c) 2018, S. Shiraiwa  
   All Rights reserved. See file COPYRIGHT for details.

   Precondirioners

   ### simple scenario,,,
   # In this mode, one choose preconditioner block from GUI
   # gui says.... A1 : ams(singular=True)

   # code in iterative_model
   g = DiagonalPrcGen(opr=opr, engin=engine, gui=gui)
   prc = g()
   ams.set_param(g, "A1")
   blk = ams(singular=True) # generate preconditioner
   prc.SetDiagonalBlock(kblockname, blk) # set to block


   ### scenario 2
   # in this mode, a user chooses operator type(diagonal, lowertriagular)
   # and block filling is done in decorated function
   # gui says.... D(*args, **kwargs)

   # code in iterative_model
   expr = self.gui.adv_prc  # expr: expression to create a generator. 
                            # (example) expr = "D('A1')"
   gen = eval(expr, self.gui._global_ns)
   gen.set_param(opr, engine, gui)
   M = gen()

   @prc.blk_diagonal (or @prc.blk_lowertriangular)
   def D(prc, g, *args, **kwargs):
       # first two argments are mandatory
       # prc: preconditioner such as mfem.BlockDiagonalPreconditioner
       # g  : preconditioner generator, which can be used to 
       #      access operator, gui, engine,,,,

       ams.set_param(g, "A1")
       blk = ams(singular=True) # generate preconditioner
       k = g.get_row_by_name("a")
       prc.SetDiagonalBlock(k, blk) # set to block
       return prc

   ### scenario 3
   # In this mode, a user has full control of preconditoner construction
   # Mult, SetOperator must defined
   # gui says.... S(*args, **kwargs)

   # code in iterative_model
   S.set_param(opr, engine, gui)
   prc = S()

   @prc.blk_generic
   def S(prc, g, *args, **kwargs):
       D.copy_param(g)
       prc1 = D()
       LT.copy_param(g)
       prc2 = LT()
       prc._prc1 = prc1
       prc._prc2 = prc2

   @S.Mult
   def S.Mult(prc, x, y):
       tmpy = mfem.Vector(); 
       prc._prc1.Mult(x, tmpy)
       prc._prc2.Mult(tmpy, y)

   @S.SetOperator
   def S.SetOperator(prc, opr):


'''
import weakref

from petram.mfem_config import use_parallel
if use_parallel:
    from petram.helper.mpi_recipes import *
    from mfem.common.parcsr_extra import *
    import mfem.par as mfem

    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    smyid = '{:0>6d}'.format(myid)
    from mfem.common.mpi_debug import nicePrint
else:
    import mfem.ser as mfem

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('Preconditioner')


class PreconditionerBlock(object):
    def __init__(self, func):
        self.func = func

    def set_param(self, prc, blockname):
        self.prc = prc
        self.blockname = blockname

    def __call__(self, *args, **kwargs):
        kwargs['prc'] = self.prc
        kwargs['blockname'] = self.blockname
        return self.func(*args, **kwargs)


class PrcCommon(object):
    def set_param(self, opr, name, engine, gui):
        self._opr = weakref.ref(opr)
        self.gui = gui
        self.name = name  # variable name on opr
        self._engine = weakref.ref(engine)

    def copy_param(self, g):
        self._opr = g._opr
        self.gui = g.gui
        self._engine = g._engine

    @property
    def engine(self):
        return self._engine()

    @property
    def opr(self):
        return self._opr()

    def get_row_by_name(self, name):
        return self.name.index(name)

    def get_col_by_name(self, name):
        return self.name.index(name)

    def get_operator_block(self, r, c):
        # if linked_op exists (= op is set from python).
        # try to get it
        # print(self.opr._linked_op)
        if hasattr(self.opr, "_linked_op"):
            try:
                return self.opr._linked_op[(r, c)]
            except KeyError:
                return None
        else:
            blk = self.opr.GetBlock(r, c)
            if use_parallel:
                return mfem.Opr2HypreParMat(blk)
            else:
                return mfem.Opr2SparseMat(blk)

    def get_diagoperator_by_name(self, name):
        r = self.get_row_by_name(name)
        c = self.get_row_by_name(name)
        return self.get_operator_block(r, c)

    def get_test_fespace(self, name):
        fes = self.engine.fespaces[name]

        return fes


class PrcGenBase(PrcCommon):
    def __init__(self, func=None, opr=None, engine=None, gui=None, name=None):
        self.func = func
        self._params = (tuple(), dict())
        self.setoperator_func = None
        if gui is not None:
            self.set_param(opr, name, engine, gui)

    def SetOperator(self, func):
        self.setoperator_func = func


class DiagonalPrcGen(PrcGenBase):
    def __call__(self, *args, **kwargs):
        offset = self.opr.RowOffsets()
        D = mfem.BlockDiagonalPreconditioner(offset)
        if self.func is not None:
            self.func(D, self, *args, **kwargs)
        return D


class LowerTriangluarPrcGen(PrcGenBase):
    def __call__(self, *args, **kwargs):
        offset = self.opr.RowOffsets()
        LT = mfem.BlockLowerTriangularPreconditioner(offset)
        if self.func is not None:
            self.func(LT, self, *args, **kwargs)
        return LT


class GenericPreconditionerGen(PrcGenBase):
    def __init__(self, func=None, opr=None, engine=None, gui=None):
        PrcGenBase.__init__(self, func=func, opr=opr, engine=engine, gui=gui)
        self.mult_func = None
        self.setoperator_func = None

    def Mult(self, func):
        self.mult_func = func

    def __call__(self,  *args, **kwargs):
        assert self.mult_func is not None, "Mult is not defined"
        assert self.setoperator_func is not None, "SetOperator is not defined"

        dargs, dkwargs = self._params
        assert len(dargs) == 0,  "Decorator allows only keyword argments"

        prc = GenericPreconditioner(self)

        for key in dkwargs:
            kwargs[key] = dkwargs[key]
        prc = self.func(prc,  *args, **kwargs)
        return prc


class _prc_decorator(object):
    def block(self, func):
        class deco(PreconditionerBlock):
            def __init__(self, func):
                self.func = func
        return deco(func)

        '''
        def dec(*args, **kwargs):
            obj = PreconditionerBlock(func)
            obj._params = (args, kwargs)
            return obj
        return dec
        '''

    def blk_diagonal(self, func):
        class deco(DiagonalPrcGen):
            def __init__(self, func):
                self.func = func
        return deco(func)

    def blk_lowertriangular(self, func):
        class deco(LowerTriangularPrcGen):
            def __init__(self, func):
                self.func = func
        return deco(func)

    def blk_generic(self, *dargs, **dkargs):
        def wrapper(func):
            class deco(GenericPreconditionerGen):
                def __init__(self, func):
                    GenericPreconditionerGen.__init__(self, func)
                    self._params = (dargs, dkargs)
                    self.func = func
            return deco(func)
        return wrapper


prc = _prc_decorator()

#
#  prc.block
#
#    in prc block, following parameters are defined in kwargs
#       prc : block preconditioner generator
#       blockname : row name in prc

#       prc knows...
#       engine : petram.engin
#       gui    : iteretavie_solver_model object
#       opr    : operator to be smoothed

SparseSmootherCls = {"Jacobi": (mfem.DSmoother, 0),
                     "l1Jacobi": (mfem.DSmoother, 1),
                     "lumpedJacobi": (mfem.DSmoother, 2),
                     "GS": (mfem.GSSmoother, 0),
                     "forwardGS": (mfem.GSSmoother, 1),
                     "backwardGS": (mfem.GSSmoother, 2), }


def _create_smoother(name, mat):
    if use_parallel:
        smoother = mfem.HypreSmoother(mat)
        smoother.SetType(getattr(mfem.HypreSmoother, name))
    else:
        cls = SparseSmootherCls[name][0]
        arg = SparseSmootherCls[name][1]
        smoother = cls(mat, arg)
        smoother.iterative_mode = False
    return smoother


def mfem_smoother(name, **kwargs):
    prc = kwargs.pop('prc')
    blockname = kwargs.pop('blockname')
    row = prc.get_row_by_name(blockname)
    col = prc.get_col_by_name(blockname)
    mat = prc.get_operator_block(row, col)

    if isinstance(mat, mfem.ComplexOperator):
        conv = mat.GetConvention()

        blockOffsets = mfem.intArray()
        blockOffsets.SetSize(3)
        blockOffsets[0] = 0
        blockOffsets[1] = mat.Height()//2
        blockOffsets[2] = mat.Height()//2
        blockOffsets.PartialSum()

        smoother = mfem.BlockDiagonalPreconditioner(blockOffsets)

        m_r = mat._real_operator
        #m_i = mat._imag_operator
        pc_r = _create_smoother(name, m_r)
        pc_i = mfem.ScaledOperator(pc_r,
                                   1 if conv == mfem.ComplexOperator.HERMITIAN else -1)

        smoother.SetDiagonalBlock(0, pc_r)
        smoother.SetDiagonalBlock(1, pc_i)
        smoother._smoothers = (pc_r, pc_i)

    else:
        smoother = _create_smoother(name, mat)
    return smoother


@prc.block
def GS(**kwargs):
    return mfem_smoother('GS', **kwargs)


@prc.block
def l1GS(**kwargs):
    return mfem_smoother('l1GS', **kwargs)


@prc.block
def l1GStr(**kwargs):
    return mfem_smoother('l1GStr', **kwargs)


@prc.block
def forwardGS(**kwargs):
    return mfem_smoother('forwardGS', **kwargs)


@prc.block
def backwardGS(**kwargs):
    return mfem_smoother('backwardGS', **kwargs)


@prc.block
def Jacobi(**kwargs):
    return mfem_smoother('Jacobi', **kwargs)


@prc.block
def l1Jacobi(**kwargs):
    return mfem_smoother('l1Jacobi', **kwargs)


@prc.block
def lumpedJacobi(**kwargs):
    return mfem_smoother('lumpedJacobi', **kwargs)


@prc.block
def Chebyshev(**kwargs):
    return mfem_smoother('Chebyshev', **kwargs)


@prc.block
def Taubin(**kwargs):
    return mfem_smoother('Taubin', **kwargs)


@prc.block
def FIR(**kwargs):
    return mfem_smoother('FIR', **kwargs)


@prc.block
def schwarz(**kwargs):
    assert use_parallel, "Schwarz smoother supports only parallel mode"

    prc = kwargs.pop('prc')
    blockname = kwargs.pop('blockname')

    fes = prc.get_test_fespace(blockname)
    fes_info = prc.engine.fespaces.get_fes_info(fes)

    use_basemesh = kwargs.pop('basemesh', -1)
    if use_basemesh == -1:
        target = kwargs.pop('ref', fes_info['refine'])
        ref_level = fes_info['refine'] - target
        fes_info['refine'] = target
        fes0 = prc.engine.fespaces.get_fes_from_info(fes_info)
        pmesh = fes0.GetParMesh()
    else:
        pmesh = prc.engine.base_meshes[fes_info['emesh_idx']]
        ref_level = use_basemesh

    #dprint1(pmesh, pmesh.GetNE(), ref_level)

    row = prc.get_row_by_name(blockname)
    col = prc.get_col_by_name(blockname)
    mat = prc.get_operator_block(row, col)

    iter = kwargs.pop('iter', 1)
    theta = kwargs.pop('theta', 1)

    if isinstance(mat, mfem.ComplexOperator):
        conv = mat.GetConvention()

        m_r = mat._real_operator    # HypreParMatrix
        m_i = mat._imag_operator    # HypreParMatrix

        s = fes.GlobalTrueVSize()
        AZ = mfem.ComplexHypreParMatrix(m_r, m_i, False, False, conv)
        #AZ = mfem.ComplexHypreParMatrix(m_r, None, False, False, conv)
        M = mfem.ComplexSchwarzSmoother(pmesh, ref_level, fes, AZ)
        M.SetDumpingParam(theta)
        M.SetNumSmoothSteps(iter)
        M._linked_obj = (pmesh, fes, AZ)

    else:
        M = mfem.SchwarzSmoother(pmesh, ref_level, fes, mat)
        M._linked_obj = (pmesh, fes, mat)

    return M


@prc.block
def ams(singular=False, **kwargs):
    prc = kwargs.pop('prc')
    blockname = kwargs.pop('blockname')
    print_level = kwargs.pop('print_level', -1)

    row = prc.get_row_by_name(blockname)
    col = prc.get_col_by_name(blockname)
    mat = prc.get_operator_block(row, col)
    fes = prc.get_test_fespace(blockname)
    inv_ams = mfem.HypreAMS(mat, fes)
    if singular:
        inv_ams.SetSingularProblem()
    inv_ams.SetPrintLevel(print_level)
    inv_ams.iterative_mode = False
    return inv_ams


@prc.block
def boomerAMG(**kwargs):
    prc = kwargs.pop('prc')
    blockname = kwargs.pop('blockname')
    print_level = kwargs.pop('print_level', -1)

    row = prc.get_row_by_name(blockname)
    col = prc.get_col_by_name(blockname)
    mat = prc.get_operator_block(row, col)

    inv_boomeramg = mfem.HypreBoomerAMG(mat)
    inv_boomeramg.SetPrintLevel(print_level)
    inv_boomeramg.iterative_mode = False

    return inv_boomeramg


@prc.block
def schur(*names, **kwargs):
    # schur("A1", "B1", scale=(1.0, 1e3))
    prc = kwargs.pop('prc')
    blockname = kwargs.pop('blockname')

    r0 = prc.get_row_by_name(blockname)
    c0 = prc.get_col_by_name(blockname)

    scales = kwargs.pop('scale', [1]*len(names))
    print_level = kwargs.pop('print_level', -1)

    S = []
    for name, scale in zip(names, scales):
        r1 = prc.get_row_by_name(name)
        c1 = prc.get_col_by_name(name)
        B = prc.get_operator_block(r0, c1)
        Bt = prc.get_operator_block(r1, c0)

        B0 = prc.get_operator_block(r1, c1)
        if use_parallel:
            Bt = Bt.Transpose()
            Bt = Bt.Transpose()
            Md = mfem.HypreParVector(MPI.COMM_WORLD,
                                     B0.GetGlobalNumRows(),
                                     B0.GetColStarts())
        else:
            Bt = Bt.Copy()
            Md = mfem.Vector()
        B0.GetDiag(Md)
        Md *= scale
        if use_parallel:

            Bt.InvScaleRows(Md)
            S.append(mfem.ParMult(B, Bt))
        else:
            S.append(mfem.Mult(B, Bt))

    if use_parallel:
        from mfem.common.parcsr_extra import ToHypreParCSR, ToScipyCoo

        S2 = [ToScipyCoo(s) for s in S]
        for s in S2[1:]:
            S2[0] = S2[0]+s
        S = ToHypreParCSR(S2[0].tocsr())
        invA0 = mfem.HypreBoomerAMG(S)

    else:
        from mfem.common.sparse_utils import sparsemat_to_scipycsr

        S2 = [sparsemat_to_scipycsr(s).tocoo() for s in S]
        for s in S2[1:]:
            S2[0] = S2[0]+s
        S = mfem.SparseMatrix(S2.tocsr())
        invA0 = mfem.DSmoother(S)

    invA0.iterative_mode = False
    invA0.SetPrintLevel(print_level)
    invA0._S = S

    return invA0


@prc.block
def mumps(guiname, **kwargs):
    # mumps("mumps1")
    from petram.solver.mumps_model import MUMPSPreconditioner
    prc = kwargs.pop('prc')
    blockname = kwargs.pop('blockname')
    kwargs = {}
    if "silent" in kwargs:
        kwargs['silent'] = kwargs.pop('silent')
    r0 = prc.get_row_by_name(blockname)
    c0 = prc.get_col_by_name(blockname)
    A0 = prc.get_operator_block(r0, c0)

    invA0 = MUMPSPreconditioner(A0, gui=prc.gui[guiname],
                                engine=prc.engine,
                                **kwargs)
    return invA0


@prc.block
def gmres(atol=0.0, rtol=0.0, max_num_iter=5,
          kdim=50, print_level=-1,
          preconditioner=None, **kwargs):
    prc = kwargs.pop('prc')
    blockname = kwargs.pop('blockname')

    if use_parallel:
        gmres = mfem.GMRESSolver(MPI.COMM_WORLD)
    else:
        gmres = mfem.GMRESSolver()
    gmres.iterative_mode = False
    gmres.SetRelTol(rtol)
    gmres.SetAbsTol(atol)
    gmres.SetMaxIter(max_num_iter)
    gmres.SetPrintLevel(print_level)
    gmres.SetKDim(kdim)
    r0 = prc.get_row_by_name(blockname)
    c0 = prc.get_col_by_name(blockname)

    A0 = prc.get_operator_block(r0, c0)

    gmres.SetOperator(A0)
    if preconditioner is not None:
        gmres.SetPreconditioner(preconditioner)
        # keep this object from being freed...
        gmres._prc = preconditioner
    return gmres


@prc.block
def fgmres(atol=0.0, rtol=0.0, max_num_iter=5,
           kdim=50, print_level=-1,
           preconditioner=None, **kwargs):
    prc = kwargs.pop('prc')
    blockname = kwargs.pop('blockname')

    if use_parallel:
        fgmres = mfem.FGMRESSolver(MPI.COMM_WORLD)
    else:
        fgmres = mfem.FGMRESSolver()
    fgmres.iterative_mode = False
    fgmres.SetRelTol(rtol)
    fgmres.SetAbsTol(atol)
    fgmres.SetMaxIter(max_num_iter)
    fgmres.SetPrintLevel(print_level)
    fgmres.SetKDim(kdim)
    r0 = prc.get_row_by_name(blockname)
    c0 = prc.get_col_by_name(blockname)

    A0 = prc.get_operator_block(r0, c0)

    fgmres.SetOperator(A0)
    if preconditioner is not None:
        fgmres.SetPreconditioner(preconditioner)
        # keep this object from being freed...
        fgmres._prc = preconditioner
    return fgmres


@prc.block
def pcg(atol=0.0, rtol=0.0, max_num_iter=5,
        print_level=-1, preconditioner=None, **kwargs):
    prc = kwargs.pop('prc')
    blockname = kwargs.pop('blockname')

    if use_parallel:
        pcg = mfem.CGSolver(MPI.COMM_WORLD)
    else:
        pgc = mfem.CGSolver()
    pcg.iterative_mode = False
    pcg.SetRelTol(rtol)
    pcg.SetAbsTol(atol)
    pcg.SetMaxIter(max_num_iter)
    pcg.SetPrintLevel(print_level)
    r0 = prc.get_row_by_name(blockname)
    c0 = prc.get_col_by_name(blockname)

    A0 = prc.get_operator_block(r0, c0)

    pcg.SetOperator(A0)
    if preconditioner is not None:
        pcg.SetPreconditioner(preconditioner)
        # keep this object from being freed...
        pcg._prc = preconditioner
    return pcg


@prc.block
def bicgstab(atol=0.0, rtol=0.0, max_num_iter=5,
             print_level=-1, preconditioner=None, **kwargs):
    prc = kwargs.pop('prc')
    blockname = kwargs.pop('blockname')

    if use_parallel:
        bicgstab = mfem.BiCGSTABSolver(MPI.COMM_WORLD)
    else:
        bicgstab = mfem.BiCGSTABSolver()
    bicgstab.iterative_mode = False
    bicgstab.SetRelTol(rtol)
    bicgstab.SetAbsTol(atol)
    bicgstab.SetMaxIter(max_num_iter)
    bicgstab.SetPrintLevel(print_level)
    r0 = prc.get_row_by_name(blockname)
    c0 = prc.get_col_by_name(blockname)

    A0 = prc.get_operator_block(r0, c0)

    bicgstab.SetOperator(A0)
    if preconditioner is not None:
        bicgstab.SetPreconditioner(preconditioner)
        # keep this object from being freed...
        bicgstab._prc = preconditioner
    return bicgstab


# these are here to use them in script w/o disnginguishing
# if mfem is mfem.par or mfem.ser
BlockDiagonalPreconditioner = mfem.BlockDiagonalPreconditioner
BlockLowerTriangularPreconditioner = mfem.BlockLowerTriangularPreconditioner


class GenericPreconditioner(mfem.Solver, PrcCommon):
    def __init__(self, gen):
        self.offset = gen.opr.RowOffsets()
        self.mult_func = gen.mult_func
        self.setoperator_func = gen.setoperator_func
        self.name = gen.name
        super(GenericPreconditioner, self).__init__()
        self.copy_param(gen)

    def Mult(self, *args):
        return self.mult_func(self, *args)

    def SetOperator(self, opr):
        opr = mfem.Opr2BlockOpr(opr)
        self._opr = weakref.ref(opr)
        self.offset = opr.RowOffsets()
        return self.setoperator_func(self, opr)



'''

    extract_mesh_data: 

    digest mesh object data for plotting.
    if also extend mfem:Mesh object to include the 
    extended_connectivity information
  
    extended_connectivity = {'line2vert':{},   # two edges to make an edge
                             'surf2line':{},   # line loop for surface
                             'vol2surf': {},   # surface loop for volue
                             'line2edge':{},   # line number to mfem edge element
                             'vert2vert':{}}   # vertex number to mfem vertex id


'''
import numpy as np
from collections import defaultdict

from petram.mesh.find_edges import find_edges
from petram.mesh.find_vertex import find_vertex

from petram.mfem_config import use_parallel

if use_parallel:
    from mfem.par import GlobGeometryRefiner
    import mfem.par as mfem
    from mfem.par import ParMesh
else:
    from mfem.ser import GlobGeometryRefiner
    import mfem.ser as mfem

    # dummy class
    class ParMesh(object):
        pass
    
def extract_mesh_data_1D(mesh):
    nv = mesh.GetNV()
    X = np.vstack([mesh.GetVertexArray(k) for k in range(nv)])
    if X.shape[1] == 2:
        X = np.hstack((X, np.zeros((X.shape[0],1))))
    elif X.shape[1] == 1:
        X = np.hstack((X, np.zeros((X.shape[0],2))))
        
    cells = {}
    cell_data = {}

    attrs = mesh.GetAttributeArray()
    lverts = []; lattr=[]
    for attr in np.unique(attrs):
        idx = np.where(attrs == attr)[0]
        for i in idx:
            lverts.append(mesh.GetElementVertices(i))
        lattr.extend([attr]*len(idx))
        
    cells['line'] = np.array(lverts)
    cell_data['line'] = {}
    cell_data['line']['physical'] = np.array(lattr).flatten()

    vverts = []; vattr=[]    
    for ibe in range(mesh.GetNBE()):
        battr = mesh.GetBdrAttribute(ibe)
        vattr.append(battr)
        vverts.append(mesh.GetBdrElementVertices(ibe))
    cells['vertex'] = np.array(vverts)
    cell_data['vertex'] = {}    
    cell_data['vertex']['physical'] = np.array(vattr).flatten()

    from petram.mesh.mesh_utils import get_extended_connectivity
    
    if not hasattr(mesh, 'extended_connectivity'):
        get_extended_connectivity(mesh)
    
    return X, cells, cell_data, [None, None], None

def extract_mesh_data(mesh, refine=1):
    if isinstance(mesh, ParMesh):
        assert False, "mesh data must be processed in serial"
    
    hasNodal = mesh.GetNodalFESpace() is not None    
    ndim = mesh.Dimension()

    if hasNodal and refine != 1:
       if ndim == 3:
           from petram.mesh.read_mfemmesh3 import extract_refined_mesh_data3           
           return extract_refined_mesh_data3(mesh, refine)
       elif ndim == 2:
           from petram.mesh.read_mfemmesh2 import extract_refined_mesh_data2
           return extract_refined_mesh_data2(mesh, refine)
       else:
           from petram.mesh.read_mfemmesh1 import extract_refined_mesh_data1
           return extract_refined_mesh_data1(mesh, refine)
    if ndim == 3:
        ivert0 = [mesh.GetBdrElement(i).GetVerticesArray()
                           for i in range(mesh.GetNBE())]
        attrs = mesh.GetBdrAttributeArray()
    elif ndim == 2:
        ivert0 = [mesh.GetElement(i).GetVerticesArray()
                       for i in range(mesh.GetNE())]

        attrs = mesh.GetAttributeArray()        
    else:
        return extract_mesh_data_1D(mesh)

    nvert = np.array([len(x) for x in ivert0])
    idx3 = np.where(nvert == 3)[0]
    idx4 = np.where(nvert == 4)[0]
    ivert = []; ivert3 = None; ivert4 = None
    iv3 = []; iv4 = []

    if len(idx3) != 0:
        iv3 = [ivert0[k] for k in idx3]
        ivert3 = np.vstack(iv3)
        attrs3 = attrs[idx3]
    if len(idx4) != 0:
        iv4 = [ivert0[k] for k in idx4]        
        ivert4 = np.vstack(iv4)
        attrs4 = attrs[idx4]

    tmp = np.hstack(iv3+iv4)
    u, indices = np.unique(tmp, return_inverse=True)
    
    ll3 = 3*len(idx3)
    indices3 = indices[:ll3]
    indices4 = indices[ll3:]

    # table u -> u's idx
    table = np.zeros(np.max(u)+1, dtype=int)-1
    for k, u0 in enumerate(u):
        table[u0] = k
        
    X = np.vstack([mesh.GetVertexArray(k) for k in u])
    if X.shape[1] == 2:
        X = np.hstack((X, np.zeros((X.shape[0],1))))
    elif X.shape[1] == 1:
        X = np.hstack((X, np.zeros((X.shape[0],2))))
    
    kdom = mesh.attributes.ToList()
    
    cells = {}
    cell_data = {}
    if ivert3 is not None:
        cells['triangle'] = indices3.reshape(ivert3.shape)
        cell_data['triangle'] = {}
        cell_data['triangle']['physical'] = attrs3
    if ivert4 is not None:
        cells['quad'] = indices4.reshape(ivert4.shape)
        cell_data['quad'] = {}
        cell_data['quad']['physical'] = attrs4


    from petram.mesh.mesh_utils import populate_plotdata
    l_s_loop = populate_plotdata(mesh, table, cells, cell_data)
    iedge2bb = None # is it used?    
    return X, cells, cell_data, l_s_loop, iedge2bb
    '''
    get_extended_connectivity(mesh)
    ec = mesh.extended_connectivity

    cell_data['line'] = {}
    cell_data['vertex'] = {}
    l2e = ec['line2edge']
    v2v = ec['vert2vert']

    kedge = []

    kedge = np.array(sum([[key]*len(l2e[key]) for key in l2e], [])).astype(int)
    iverts = np.vstack([mesh.GetEdgeVertices(ie)
                        for key in l2e for ie in l2e[key]])
    cells['line'] = table[iverts]
    cell_data['line']['physical'] = np.array(kedge)

    kvert = np.array([key for key in v2v]).astype(int)    
    iverts = np.array([v2v[key] for key in v2v]).astype(int)    

    cells['vertex'] = table[iverts]    
    cell_data['vertex']['physical'] = kvert
    
    if ndim == 3:
        l_s_loop = [ec['surf2line'], ec['vol2surf']]
    elif ndim == 2:
        l_s_loop = [ec['surf2line'], None]
    else:
        l_s_loop = [None, None]

    iedge2bb = None # is it used?
    '''





from __future__ import print_function

import numpy as np
import datetime
import os
import shlex
import socket
import base64
import traceback
import subprocess as sp


def communicate_with_timeout(p,
                             maxtimeout=np.inf,
                             timeout=2,
                             progdlg=None,
                             verbose=False):

    import wx
    if progdlg is not None:
        value = progdlg.GetValue()
    time = 0
    cancelled = False
    timeout_expired = False
    while True:
        #print("communicating", time)
        try:
            p.wait(timeout=timeout)
        except sp.TimeoutExpired:
            time = time + timeout
            wx.GetApp().Yield()
            if progdlg is not None:
                if progdlg:
                    flag, skipped = progdlg.Update(value)
                    cancelled = not flag
                else:
                    #print("widget deleted")
                    cancelled = True
                if cancelled:
                    p.kill()
                    return b"", b"", False, True
            if time > maxtimeout:
                timeout_expired = True
                p.kill()
                return b"", b"", True, False
            else:
                continue
        break
    outs, errs = p.communicate()
    if errs is not None:
        print(errs.decode())

    if verbose:
        if outs is not None:
            print(outs.decode())

    return outs, errs, timeout_expired, cancelled


def launch_ssh_command(model, command, verbose=True):
    param = model.param
    hosto = param.eval('host')
    host = hosto.getvar('server')
    user = hosto.getvar('user')
    opts = hosto.get_multiplex_opts()
    aopts = hosto.get_auth_opts(no_password=True)

    opts = ' '.join(opts)
    command = ("ssh -x " + opts + aopts + user +
               '@' + host + " '" + command + "'")

    if verbose:
        print("Executing on host (ssh): " + command)

    p = sp.Popen(command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    return p


def wdir_from_datetime():
    import datetime
    import socket
    txt = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S_%f")
    hostname = socket.gethostname()
    txt = txt + '_' + hostname
    return txt


def make_remote_connection(model, host):
    '''
    host = 'eofe7.mit.edu'

    '''
    import ifigure

    proj = model.get_root_parent()
    p = proj.setting.parameters

    if p.hasvar('connection'):
        c = p.eval('connection')
    else:
        c = None
    if c is None:
        base = os.path.dirname(ifigure.__file__)
        f = os.path.join(base, 'add_on', 'setting', 'module', 'connection.py')
        c = proj.setting.add_absmodule(f)
        p.setvar('connection', '='+c.get_full_path())

    objname = host.split('.')[0]
    if not c.has_child(objname):
        c.call_method('add_connection', objname)
        proj.app.proj_tree_viewer.update_widget()
        obj = c.get_child(name=objname)
        obj.setvar('server', host)
        obj.onSetting()
    else:
        obj = c.get_child(name=objname)
        obj.onSetting()
    return obj


def clean_remote_dir(model):
    param = model.param
    rwdir = param.eval('remtoe')['rwdir']
    if rwdir is None:
        return False

    host = param.eval('host')
    host.Execute('rm ' + rwdir + '/solmesh*')
    host.Execute('rm ' + rwdir + '/soli*')
    host.Execute('rm ' + rwdir + '/solr*')

    return True


def prepare_remote_dir(model, txt='', dirbase='', progdlg=None):
    model_dir = model.owndir()
    param = model.param
    if txt == '':
        txt = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S_%f")
        hostname = socket.gethostname()
        rwdir = os.path.join(dirbase, txt+'_'+hostname)
    else:
        rwdir = os.path.join(dirbase, txt)

    host = param.eval('host')
    hostname = host.getvar('server')
    user = host.getvar('user')

    #p = host.Execute('mkdir -p ' + rwdir, nowait=True)
    command = 'mkdir -p ' + rwdir
    p = launch_ssh_command(model, command)
    #command = ("ssh -x -o PasswordAuthentication=no -o PreferredAuthentications=publickey " + user+'@' + host + " '" + command + "'")
    #p= sp.Popen(command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)

    ret = communicate_with_timeout(p, timeout=2, progdlg=progdlg)
    _outs, _errs, timeout, cancelled = ret

    if timeout or cancelled:
        return True

    param.eval('remote')['rwdir'] = rwdir
    return False


def send_file(model, progdlg, skip_mesh=False, subdir=''):
    model_dir = model.owndir()
    param = model.param

    remote = param.eval('remote')
    host = param.eval('host')
    sol = param.eval('sol')
    sol_dir = sol.owndir()

    rwdir = remote['rwdir']
    if subdir != '':
        rwdir = rwdir+'/'+subdir
    mfem_model = param.eval('mfem_model')

    p = host.PutFile(os.path.join(sol_dir, 'model.pmfm'),
                     rwdir + '/model.pmfm',
                     nowait=True)
    ret = communicate_with_timeout(p, timeout=2, progdlg=progdlg)
    _outs, _errs, timeout, cancelled = ret
    if timeout or cancelled:
        return True

    if skip_mesh:
        return False
    for od in mfem_model.walk():
        if not od.is_enabled():
            continue
        if hasattr(od, 'use_relative_path'):
            path = od.get_real_path()
            dpath = rwdir+'/'+os.path.basename(od.path)
            p = host.PutFile(path, dpath, nowait=True)
            ret = communicate_with_timeout(p, timeout=2, progdlg=progdlg)
            _outs, _errs, timeout, cancelled = ret
            if timeout or cancelled:
                return True

    return False


def retrieve_files(model, rhs=False, matrix=False, sol_dir=None):
    model_dir = model.owndir()
    param = model.param
    if sol_dir is None:
        sol_dir = model.owndir()

    def get_files(host, key):
        xx = host.Execute('ls ' + os.path.join(rwdir, key+'*')
                          ).stdout.readlines()

        files = [x.strip() for x in xx if x.find(key) != -1]
        host.GetFiles(files, sol_dir)

        xx = host.Execute('ls -d ' + os.path.join(rwdir,
                                                  'case*')).stdout.readlines()
        print(xx)
        for x in xx:
            x0 = os.path.basename(x.strip())
            if not x0.startswith('case'):
                continue
            try:
                long(x0[4:])
            except:
                continue
            print('testing', os.path.join(sol_dir, x0))
            if not os.path.exists(os.path.join(sol_dir, x0)):
                os.mkdir(os.path.join(sol_dir, x0))
            yy = host.Execute('ls ' + os.path.join(rwdir, x0,
                                                   key+'*')).stdout.readlines()
            print('!!!!!!!!!!', yy)
            files = [os.path.join(x.strip(), os.path.basename(y.strip()))
                     for y in yy if y.find(key) != -1]
            host.GetFiles(files, os.path.join(sol_dir, x0))

    import os

    host = param.eval('host')
    remote = param.eval('remote')
    rwdir = remote['rwdir']

    get_files(host, 'solr')
    get_files(host, 'soli')
    get_files(host, 'solmesh')
    if matrix:
        get_files(host, 'matrix')
    if rhs:
        get_files(host, 'rhs')


def _get_job_queue(model, command,  host=None, user=None, progdlg=None, configext=''):
    '''
    param = model.param
    hosto = param.eval('host')
    host = hosto.getvar('server')
    user = hosto.getvar('user')
    '''
    # command = ("ssh -o PasswordAuthentication=no -o PreferredAuthentications=publickey " +
    #           user+'@' + host + " 'cat $PetraM/etc/queue_config'" )
    #p= sp.Popen(command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)

    p = launch_ssh_command(model, command)
    ret = communicate_with_timeout(p, maxtimeout=30,
                                   timeout=2, progdlg=progdlg)
    outs, errs, timeout, cancelled = ret

    if timeout:
        return False, None
    if cancelled:
        return False, None

    lines = [x.strip() for x in outs.decode('utf-8').split('\n')]

    try:
        value = interpret_job_queue_file(lines)
    except BaseException:
        traceback.print_exc()
        return False, None
    return True, value


def get_job_queue(model, host=None, user=None, progdlg=None, configext=''):

    # try this first
    command = '$PetraM/bin/get_job_queues.sh'+configext
    ret = _get_job_queue(model, command,  host=host,
                         user=user, progdlg=progdlg, configext=configext)
    if ret[0]:
        return ret

    # try this (old method)
    command = 'cat $PetraM/etc/queue_config'+configext
    ret = _get_job_queue(model, command,  host=host,
                         user=user, progdlg=progdlg, configext=configext)
    return ret


def interpret_job_queue_file(lines):
    lines = [x.strip() for x in lines if not x.startswith("#")
             if len(x.strip()) != 0]

    if len(lines) == 0:
        return None

    q = {'type': lines[0], 'queues': [],
         'scratch': "~/myscratch", "notice": [], "versions": {}}
    for l in lines[1:]:
        if l.startswith('KEYWORD'):
            if not 'keywords' in q:
                q['keywords'] = []
            q['keywords'].append(l.split(':')[1])
        elif l.startswith('QUEUE'):
            q['queues'].append({'name': l.split(':')[1]})
        elif l.startswith('VERSION'):
            q['versions'][l.split(':')[1]] = []
        elif l.startswith('VEROPT'):
            version = l.split(':')[1]
            opt = l.split(':')[2]
            param = ':'.join(l.split(':')[3:])
            q['versions'][version].append((opt, param))
        elif l.startswith('SCRATCH'):
            q['scratch'] = l.split(':')[1].rstrip()
        elif l.startswith('NOTICE'):
            q['notice'].append((":".join(l.split(':')[1:])).rstrip())
        else:
            data = ':'.join(l.split(':')[1:])
            param = l.split(':')[0]
            q['queues'][-1][param] = data

    return q


def submit_job(model, progdlg=None, sh_command="$PetraM/bin/launch_petram.sh"):
    param = model.param
    host = param.eval('host')
    remote = param.eval('remote')
    rwdir = remote['rwdir']
    hostname = host.getvar('server')
    user = host.getvar('user')
    # p= sp.Popen("ssh " + user+'@' + hostname + " 'printf $PetraM'",
    #              shell=True, stdout=sp.PIPE)
    #PetraM = p.stdout.readlines()[0].decode('utf-8').strip()

    w = remote["walltime"]
    n = str(remote["num_cores"])
    N = str(remote["num_nodes"])
    o = str(remote["num_openmp"])
    q = str(remote["queue"])

    adv = str(remote["adv_opts"])
    adv = "\n".join([x.strip()
                    for x in adv.split("\n") if not x.strip().startswith("#")])
    env = str(remote["env_opts"])
    env = "\n".join([x.strip()
                    for x in env.split("\n") if not x.strip().startswith("#")])

    lk = []
    for k, v in remote["log_keywords"]:
        if v:
            lk.append(k.strip())
    lk = ','.join(lk)

    lt = str(remote["log_txt"])
    lt = "'".join(lt.split('"'))

    nt = str(remote["notification"])

    lk = base64.b64encode(lk.encode()).decode()
    lt = base64.b64encode(lt.encode()).decode()
    env = base64.b64encode(env.encode()).decode()
    adv = base64.b64encode(adv.encode()).decode()

    q1 = q.strip().split("(")[0]
    q2 = "" if q.find("(") == -1 else "(".join(q.strip().split("(")[1:])[:-1]
    # replace short/dev -> short_dev
    if q2 != "":
        q2 = "_".join(q2.split("/"))

    exe = (sh_command + ' -N ' + N + ' -P ' + n + ' -W ' + w + ' -O ' + o + ' -Q ' + q1
           + ' -L ' + lt + ' -K ' + lk + ' -M ' + nt)
    if q2 != "":
        exe = exe + ' -V ' + q2
    if adv != "":
        exe = exe + ' -A ' + adv
    if env != "":
        exe = exe + ' -E ' + env

    command = 'cd '+rwdir+';'+exe
    p = launch_ssh_command(model, command)
    ret = communicate_with_timeout(p, timeout=2, progdlg=progdlg, verbose=True)
    _outs, _errs, timeout, cancelled = ret

    if timeout or cancelled:
        return True
    return False



import wx
import wx.grid as gr


class NonlinearTermPanel0(wx.Panel):
    def __init__(self, parent, id, names = None):
        wx.Panel.__init__(self, parent, id)

        self.names = [] if names is None else names
        self.grids = {}
        self.btns = {}
        
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        for n in names:
            self.add_panel(n, sizer)

    def add_panel(self, name, top_sizer):
        sizer = wx.BoxSizer(wx.VERTICAL)
        top_sizer.Add(sizer, 0, wx.EXPAND)        
        h = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(h, 0, wx.EXPAND)
        t=wx.StaticText(self, wx.ID_ANY, "  " + name + "  " )
        bp = wx.Button(self, wx.ID_ANY, '+', style=wx.BU_EXACTFIT)
        bm = wx.Button(self, wx.ID_ANY, '-', style=wx.BU_EXACTFIT)

        h.Add(t, 0)
        h.AddStretchSpacer()        
        h.Add(bp, 0)
        h.Add(bm, 0)
        def hit_plus(evt, name=name):
            self.hit_plus(evt, name)
        def hit_minus(evt, name=name):
            self.hit_minus(evt, name)
        self.Bind(wx.EVT_BUTTON, hit_plus, bp)
        self.Bind(wx.EVT_BUTTON, hit_minus, bm)

        grid = gr.Grid(self, -1)
        grid.CreateGrid(3, 2)
        grid.HideRowLabels()
        grid.SetColLabelValue(0, 'variable')
        grid.SetColLabelValue(1, 'expression')        
        sizer.Add(grid, 1, wx.EXPAND|wx.ALL, 5)

        def onSize(evt, grid = grid):
            from ifigure.utils.wx3to4 import grid_ClientSizeTuple
            width, height = grid_ClientSizeTuple(grid)
            s = width - grid.GetColSize(0)
            if s < 1: return
            grid.SetColSize(1, s)
        
        grid.Bind(wx.EVT_SIZE, onSize)
        
        self.grids[name] = grid
        self.btns[name] = (bp, bm)
        
    def hit_plus(self, evt, name):
        grid = self.grids[name]
        grid.AppendRows(1)
        self.Layout()
        self.send_event(evt)        
        
    def hit_minus(self, evt, name):
        grid = self.grids[name]
        cursor = grid.GetGridCursorRow()
        grid.DeleteRows(cursor)
        self.Layout()
        self.send_event(evt)

    def SetValue(self, value):
        for name in self.names:
            grid = self.grids[name]
            for k in range(grid.GetNumberRows()):
                grid.SetCellValue(k, 0, '')
                grid.SetCellValue(k, 1, '')                
            k = 0
            if not name in value: continue
            for c1, c2 in value[name]:
                grid.SetCellValue(k, 0, c1)
                grid.SetCellValue(k, 1, c2)                
                k = k + 1

    def GetValue(self):
        value = {}
        for name in self.names:
            grid = self.grids[name]
            v = []
            for k in range(grid.GetNumberRows()):
                a = grid.GetCellValue(k, 0)
                b = grid.GetCellValue(k, 1)
                if a.strip() == '': continue
                v.append((a, b))
            value[name] = v
        return value

    def Enable(self, value):
        for name in self.names:
            grid = self.grids[name]
            grid.Enable(value)
            self.btns[name][0].Enable(value)
            self.btns[name][1].Enable(value)
            
    def send_event(self, evt):
        evt.SetEventObject(self)
        if hasattr(self.GetParent(), 'send_event'):
            self.GetParent().send_event(self, evt)
        
class NonlinearTermPanel(wx.Panel):
    def __init__(self, parent, id, setting = None):
        setting = {} if setting is None else setting
        names = setting.pop('names', [])
        wx.Panel.__init__(self, parent, id)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        
        self.bt_use_nl = wx.CheckBox(self, wx.ID_ANY, 'Use NL')
        self.nl_terms =  NonlinearTermPanel0(self, wx.ID_ANY, names = names)
        
        sizer.Add(self.bt_use_nl, 0, wx.EXPAND)
        sizer.Add(self.nl_terms, 1 , wx.EXPAND|wx.ALL, 3)
        
        self.bt_use_nl.Bind(wx.EVT_CHECKBOX, self.onHitToggle)
        self.Bind(gr.EVT_GRID_CELL_CHANGED, self.onCellChanged)        
        self.Layout()
        
    def SetValue(self, value):
        self.bt_use_nl.SetValue(value[0])
        self.nl_terms.Enable(True)
        self.nl_terms.SetValue(value[1])
        if value[0]:
            self.nl_terms.Enable(True)
        else:
            self.nl_terms.Enable(False)            

    def GetValue(self):
        return [self.bt_use_nl.GetValue(),
                self.nl_terms.GetValue()]

    def onHitToggle(self, evt):
        tg =self.bt_use_nl.GetValue()
        if tg:
            self.nl_terms.Enable(True)
        else:
            self.nl_terms.Enable(False)
        self.send_event(self, evt)
        
    def onCellChanged(self, evt):
        self.send_event(self, evt)

    def send_event(self, obj, evt):
        #print("sending event", self.GetValue())
        evt.SetEventObject(self)
        if hasattr(self.GetParent(), 'send_event'):
            self.GetParent().send_event(self, evt)

class Example(wx.Frame):
    def __init__(self, parent=None, title='test'):
        super(Example, self).__init__(parent, title=title)
        w = NonlinearTermPanel(self, wx.ID_ANY, setting = {'names':['k', 'alpha']})
                 
        self.SetSizer(wx.BoxSizer(wx.VERTICAL))
        self.GetSizer().Add(w, 1, wx.EXPAND, 0)

        value = [True, {'k':[('x', 'a*3'), ('y', 'b+b')], 'alpha':[]}]
        w.SetValue(value)

        self.Show()

if __name__=='__main__':
   app = wx.App(False)
   e = Example()
   app.MainLoop()                 



from __future__ import print_function

import os
import numpy as np
import wx
from collections import OrderedDict
import traceback
from ifigure.utils.cbook import BuildPopUpMenu
from ifigure.utils.edit_list import EditListPanel, ScrolledEditListPanel
from ifigure.utils.edit_list import EDITLIST_CHANGED,  EDITLIST_CHANGING
from ifigure.utils.edit_list import EDITLIST_SETFOCUS
from ifigure.widgets.miniframe_with_windowlist import MiniFrameWithWindowList
from ifigure.widgets.miniframe_with_windowlist import DialogWithWindowList

import petram
import petram.helper.pickle_wrapper as pickle

try:
    import treemixin 
except ImportError:
    from wx.lib.mixins import treemixin

from petram.mfem_model import MFEM_ModelRoot

from ifigure.ifigure_config import rcdir
petram_model_scratch=os.path.join(rcdir, 'petram_model_scratch')

from petram.utils import get_pkg_datafile
#import petram.geom

fdot = get_pkg_datafile(petram.pi, 'icon',  'dot.png')
fedge = get_pkg_datafile(petram.pi, 'icon', 'line.png')
fface = get_pkg_datafile(petram.pi, 'icon', 'face.png')
fdom = get_pkg_datafile(petram.pi, 'icon', 'domain.png')

im1 = wx.Image(fdom, wx.BITMAP_TYPE_PNG)
image1 = im1.ConvertToBitmap()
im2 = wx.Image(fface, wx.BITMAP_TYPE_PNG)
image2 = im2.ConvertToBitmap()
im3 = wx.Image(fedge, wx.BITMAP_TYPE_PNG)
image3 = im3.ConvertToBitmap()
im4 = wx.Image(fdot, wx.BITMAP_TYPE_PNG)
image4 = im4.ConvertToBitmap()

w,h = image4.GetSize()
il = wx.ImageList(w, h)
il.Add(image1)
il.Add(image2)
il.Add(image3)
il.Add(image4)

from petram.pi.simple_frame_plus import SimpleFramePlus

class EntityTree(treemixin.VirtualTree, wx.TreeCtrl):
    def __init__(self, *args, **kwargs):
        self.topwindow = kwargs.pop('topwindow')
        self.entitydim = kwargs.pop('entitydim')
        self.d= {}
        self.keys = []
        super(EntityTree, self).__init__(*args, **kwargs)
        
    def OnGetItemText(self, indices):
        if len(indices) == 1:
            txt = str(self.keys[indices[0]])
        elif len(indices) == 2:
            l = sorted(self.d[self.keys[indices[0]]])
            txt =  str(l[indices[1]])
        else:
            txt = ''
        return txt
    
    def OnGetItemTextColour(self, indices):
        return wx.BLACK
        
    def OnGetItemFont(self, indices):
        return wx.NORMAL_FONT        
        
    def OnGetChildrenCount(self, indices):
        if len(indices) == 0:
            return len(self.keys)
        elif len(indices) == 1:
            return len(self.d[self.keys[indices[0]]])
        else:
            return 0
        return 0

    def GetSelection(self):
        ## this returns only one selection
        ## called when only one element is assumed to be selected
        ret = self.GetSelections()
        if len(ret) == 0: return None
        return ret[0]
    
    def isMultipleSelection(self):
        return len(self.GetSelections()) > 1

    def set_dict(self, d):
        if d is None:
            d = {}
        self.d = d
        self.keys = sorted(list(d))
        
class SelectionPalette(SimpleFramePlus):
    def __init__(self, parent, id, title, model = None):
                       
        '''
        (use this style if miniframe is used)
        style=(wx.CAPTION|
                       wx.CLOSE_BOX|
                       wx.MINIMIZE_BOX| 
                       wx.RESIZE_BORDER|
                       wx.FRAME_FLOAT_ON_PARENT)
        '''
        style =  wx.CAPTION|wx.RESIZE_BORDER|wx.SYSTEM_MENU
        style = (wx.CAPTION|
                 wx.CLOSE_BOX|
                 wx.MINIMIZE_BOX| 
                 wx.RESIZE_BORDER|
                 wx.FRAME_FLOAT_ON_PARENT|
                 wx.FRAME_TOOL_WINDOW)
        
        #style = wx.RESIZE_BORDER
        super(SelectionPalette, self).__init__(parent, id, title, style=style)
        
        #self.tree.SetSizeHints(150, -1, maxW=150)
        self.nb = wx.Notebook(self)
        self.nb.SetImageList(il)
        
        self.SetSizer(wx.BoxSizer(wx.VERTICAL))
        self.GetSizer().Add(self.nb,  1, wx.EXPAND|wx.ALL, 1)
        self.cb_update = wx.CheckBox(self, wx.ID_ANY, 'Update Figure')
        self.GetSizer().Add(self.cb_update,  0, wx.EXPAND|wx.ALL, 5)        
        
        self.p1 = wx.Panel(self.nb)
        self.p2 = wx.Panel(self.nb)
        self.p3 = wx.Panel(self.nb)
        self.p4 = wx.Panel(self.nb)
        
        self.p1.SetSizer(wx.BoxSizer(wx.HORIZONTAL))
        self.p2.SetSizer(wx.BoxSizer(wx.HORIZONTAL))
        self.p3.SetSizer(wx.BoxSizer(wx.HORIZONTAL))
        self.p4.SetSizer(wx.BoxSizer(wx.HORIZONTAL))        
        
        self.p1sizer = wx.BoxSizer(wx.VERTICAL)
        self.p2sizer = wx.BoxSizer(wx.VERTICAL)
        self.p3sizer = wx.BoxSizer(wx.VERTICAL)
        self.p4sizer = wx.BoxSizer(wx.VERTICAL)        
 
        self.p1.GetSizer().Add(self.p1sizer, 1, wx.EXPAND)
        self.p2.GetSizer().Add(self.p2sizer, 1, wx.EXPAND)
        self.p3.GetSizer().Add(self.p3sizer, 1, wx.EXPAND)
        self.p4.GetSizer().Add(self.p4sizer, 1, wx.EXPAND)

        self.nb.AddPage(self.p1, "", imageId=0)
        self.nb.AddPage(self.p2, "", imageId=1)
        self.nb.AddPage(self.p3, "", imageId=2)
        self.nb.AddPage(self.p4, "", imageId=3)
        
        self.tree1 = EntityTree(self.p1, topwindow = self, entitydim=3,
                            style = wx.TR_DEFAULT_STYLE|wx.TR_MULTIPLE)
        self.tree2 = EntityTree(self.p2, topwindow = self, entitydim=2,
                            style = wx.TR_DEFAULT_STYLE|wx.TR_MULTIPLE)
        self.tree3 = EntityTree(self.p3, topwindow = self, entitydim=1,
                            style = wx.TR_DEFAULT_STYLE|wx.TR_MULTIPLE)
        self.tree4 = EntityTree(self.p4, topwindow = self, entitydim=0,
                            style = wx.TR_DEFAULT_STYLE|wx.TR_MULTIPLE)

        self.p1sizer.Add(self.tree1, 1, wx.EXPAND)
        self.p2sizer.Add(self.tree2, 1, wx.EXPAND)
        self.p3sizer.Add(self.tree3, 1, wx.EXPAND)
        self.p4sizer.Add(self.tree4, 1, wx.EXPAND)

        self.tree1.Bind(wx.EVT_CONTEXT_MENU,
                  self.OnItemRightClick0)
        self.tree2.Bind(wx.EVT_CONTEXT_MENU,
                  self.OnItemRightClick0)
        self.tree3.Bind(wx.EVT_CONTEXT_MENU,
                  self.OnItemRightClick0)
        self.tree4.Bind(wx.EVT_CONTEXT_MENU,
                  self.OnItemRightClick0)
        
        self.Bind(wx.EVT_TREE_ITEM_RIGHT_CLICK, 
                  self.OnItemRightClick)
        self.Bind(wx.EVT_TREE_SEL_CHANGED, 
                  self.OnItemSelChanged)
        self.Bind(wx.EVT_TREE_SEL_CHANGING, 
                  self.OnItemSelChanging)
        self.Bind(wx.EVT_TREE_KEY_DOWN,
                  self.OnKeyDown)
        self.Bind(wx.EVT_KEY_UP,self.OnKeyUp)
        self.tree1.Bind(wx.EVT_KEY_UP,self.OnKeyUp)
        self.tree2.Bind(wx.EVT_KEY_UP,self.OnKeyUp)
        self.tree3.Bind(wx.EVT_KEY_UP,self.OnKeyUp)
        self.tree4.Bind(wx.EVT_KEY_UP,self.OnKeyUp)

        self.bind_sel_events()
        
        #s.Add(self.tree, 0, wx.EXPAND|wx.ALL, 1)
        #s2.Add(self.nb, 1, wx.EXPAND|wx.ALL, 1)        
        wx.GetApp().add_palette(self)
        self.Layout()
        wx.CallAfter(self.OnRefreshTree)
     
        self.panels = {}
        self.Bind(wx.EVT_CLOSE, self.OnClose)
        self.Bind(wx.EVT_CHILD_FOCUS, self.OnChildFocus)        
        self._focus_idx = None
        self._focus_obj = None
        self._copied_item = None
        self.SetSize((200,400))
        self.trees = [self.tree1, self.tree2, self.tree3, self.tree4]
        self.loops = (None, None, None, None)  #backup
        self._old_item = None
        self._key_down = (-1, False)
        self._do_veto_changing = True
        self._changed_proccessed = True        
        self._current_selection = []
        self._selected_items = []

    def get_item_from_indices(self, ktree, indices):
        tree = self.trees[ktree]
        parent = tree.GetRootItem()
        for c in indices:
           item, hoge = tree.GetFirstChild(parent)
           if not item.IsOk(): return None
           for i in range(c):
               item = tree.GetNextSibling(item)
               if not item.IsOk(): return None               
           parent = item

        return parent

    def select_by_indices(self, ktree, indices):
        tree = self.trees[ktree]
        tree.UnselectAll()
        items = tree.GetSelections()
        cindices = [tree.GetIndexOfItem(ii) for ii in items]
        for i in cindices:
            if not i in indices:            
               item = self.get_item_from_indices(ktree, i)
               if item is not None:
                   tree.SelectItem(item, select = False)
        for i in indices:
            if not i in cindices:            
                item = self.get_item_from_indices(ktree, i)
                if item is not None:
                    tree.SelectItem(item)
                

    def CallClose(self, evt):
        self.Close()

    def OnChildFocus(self, evt):
        self.GetParent()._palette_focus = 'selection'
        self.OnRefreshTree()
        evt.Skip()

    def OnItemRightClick0(self, e):
        menus = [('Refresh', self.OnRefreshTree, None)]           
        m  = wx.Menu()
        BuildPopUpMenu(m, menus, eventobj=self)
        self.PopupMenu(m, 
                       e.GetPoint())
        m.Destroy()
        
    def OnItemRightClick(self, e):
        tree = self.trees[self.nb.GetSelection()]
        menus = []        
        if tree.isMultipleSelection():
            items = tree.GetSelections()
            indices = [tree.GetIndexOfItem(ii) for ii in items]
        else:    
            indices = tree.GetIndexOfItem(tree.GetSelection())
        menus = menus + [('Refresh', self.OnRefreshTree, None)]           
        m  = wx.Menu()
        BuildPopUpMenu(m, menus, eventobj=self)
        self.PopupMenu(m, 
                       e.GetPoint())
        m.Destroy()

    def OnExportToShell(self, evt):
        indices = self.tree.GetIndexOfItem(self.tree.GetSelection())
        mm = self.model.GetItem(indices)
 
        import wx
        import ifigure.widgets.dialog as dialog
        
        app = wx.GetApp().TopWindow
        app.shell.lvar[mm.name()] = mm
        app.shell.SendShellEnterEvent()
        ret=dialog.message(app, mm.name() + ' is exported', 'Export', 0)
        
        
    def OnRefreshTree(self, evt=None):
        v = self.GetParent()

        if not v._view_mode in v._s_v_loop:
            return 
        loops = v._s_v_loop[v._view_mode]
        #loops = v._figure_data[v._view_mode]        
        if (self.loops[0] is loops[0] and
            self.loops[1] is loops[1]):
            return

        
        self.tree1.set_dict(loops[1])
        self.tree2.set_dict(loops[0])
        if self.GetParent().book.page1.axes1.has_child('edge'):
            edges = np.unique(self.GetParent().book.page1.axes1.edge.getvar('array_idx'))
        else:
            edges = []
        d = {x:[] for x in edges}

        if self.GetParent().book.page1.axes1.has_child('point'):        
            points = self.GetParent().book.page1.axes1.point.getvar('array_idx')
        else:
            points = []
        d2 = {x:[] for x in points}
        
        self.loops = (loops[0], loops[1], d, d2)
        
        self.tree3.set_dict(d)
        self.tree4.set_dict(d2)        

        self.tree1.RefreshItems()
        self.tree2.RefreshItems()
        self.tree3.RefreshItems() 
        self.tree4.RefreshItems()       
        if evt is not None: evt.Skip()

    def bind_sel_events(self):
        self.Bind(wx.EVT_TREE_SEL_CHANGED, 
                  self.OnItemSelChanged)
        self.Bind(wx.EVT_TREE_SEL_CHANGING, 
                  self.OnItemSelChanging)
    def unbind_sel_events(self):
        self.Unbind(wx.EVT_TREE_SEL_CHANGED)
        self.Unbind(wx.EVT_TREE_SEL_CHANGING)
        
    def OnKeyUp(self, evt):
        self._key_down = (-1, False)
        
    def OnKeyDown(self, evt):
        # check if alt key is pushed
        sel = self.nb.GetSelection()
        if evt.GetKeyEvent().GetKeyCode() == wx.WXK_ALT:
            self._key_down = (sel, True)
        else:
            self._key_down = (-1, False)

    def OnItemSelChanging(self, evt):
        #print("event changing")
        sel = self.nb.GetSelection()
        tree = self.trees[sel]
        
        if not evt.GetOldItem().IsOk():
            evt.Skip()
            return
        if not evt.GetItem().IsOk():
            evt.Skip()
            return 
        
        if (tree.GetIndexOfItem(evt.GetOldItem()) ==
            tree.GetIndexOfItem(evt.GetItem())):
           #print("calling Veto")
           evt.Veto()

    
    def OnItemSelChanged(self, evt = None):
        #print("event changed")
        ktree = self.nb.GetSelection()
        tree = self.trees[ktree]
        self._do_veto_changing = True
        self._changed_proccessed = True

        if tree.GetSelection() is None: return
        
        if (self._key_down[0] == ktree) and self._key_down[1]:
            #print("check1 ", self._selected)
            item = tree.GetIndexOfItem(evt.GetItem())
            #print("check2 ", item)
            if not (item in self._selected):
                new_selected = self._selected + [item]
                #print("calling new selection:", new_selected)
                wx.CallAfter(self.select_by_indices, ktree, new_selected)
                self._selected = new_selected
                evt.Skip()
                return
        else:
            items = tree.GetSelections()
            indices = [tree.GetIndexOfItem(ii) for ii in items]
            self._selected = indices
            
        #print("items in changed", self._selected)
        
        '''
        if ((self._key_down[0] == sel) and self._key_down[1] and
            not self._do_vete_changing):
            #self.unbind_sel_events()
            if evt.GetItem() is evt.GetOldItem():
                print("removing ",evt.GetOldItem())
                tree.SelectItem(evt.GetOldItem(), select = False)
                #evt.Veto()                
            else:
                print("adding ", tree.GetIndexOfItem(self._old_item))
                tree.SelectItem(self._old_item)
                #evt.Veto()
            #wx.CallAfter(self.bind_sel_events)
            self._do_vete_changing = True
            print("1")
        else:
            print("2")            
            self._do_vete_changing = False

        '''

        if self.cb_update.GetValue():
            v= [];s=[];l=[];p=[]
            if ktree == 0:
                d = self.loops[1]
                keys = sorted(list(d))
                for i in self._selected:
                    if len(i) == 1:
                        v.append(keys[i[0]])
                    else:
                        s.append(sorted(d[keys[i[0]]])[i[1]])
            elif ktree == 1:
                d = self.loops[0]
                keys = sorted(list(d))
                for i in self._selected:
                    if len(i) == 1:
                        s.append(keys[i[0]])
                    else:
                        l.append(sorted(d[keys[i[0]]])[i[1]])
            elif ktree == 2:
                d = self.loops[2]
                keys = sorted(list(d))
                for i in self._selected:
                    if len(i) == 1:
                        l.append(keys[i[0]])
                    else:
                        p.append(sorted(d[keys[i[0]]])[i[1]])
            elif ktree == 3:
                d = self.loops[3]
                keys = sorted(list(d))
                for i in self._selected:
                    if len(i) == 1:
                        p.append(keys[i[0]])
                    else:
                        pass
            else:
                pass

            vv = self.GetParent()
            if len(v) > 0:
                vv.change_panel_button('domain')
                vv.highlight_domain(v)
                vv._dom_bdr_sel = (v, [], [], [])
                
            elif len(s) > 0:
                vv.change_panel_button('face')                
                vv.highlight_face(s)
                vv._dom_bdr_sel = ([], s, [], [])
                
            elif len(l) > 0:
                vv.change_panel_button('edge')
                vv.highlight_edge(l)
                vv._dom_bdr_sel = ([], [], l, [])
                
            elif len(p) > 0:
                vv.change_panel_button('dot')                
                vv.highlight_point(p)
                vv._dom_bdr_sel = ([], [], [], p)
                
            else:
                vv.highlight_none()
                vv._dom_bdr_sel = ([], [], [], [])
                
        if evt is not None:
            evt.Skip()
    
    def OnClose(self, evt):
        wx.GetApp().rm_palette(self)
        self.GetParent().selection_palette = None
        evt.Skip()
        

        

        
        




import numpy as np
from numba import njit, float64, int32, types
from numba.types import Array

arr1D = Array(float64, 1, 'A', readonly=True)
arr2D = Array(float64, 2, 'A', readonly=True)
arr3D = Array(float64, 3, 'A', readonly=True)


@njit(float64(arr1D, arr1D, float64))
def interp1d_nearest(x, p, x0):
    '''
    1D nearest neighbor interpolation
       x 1D array. This must be monotocically increasing. Note the routine does not check this.
       p 1D array
       x0 point to interpolate
    '''

    i = np.argmin(np.abs(x - x0))
    return p[i]


@njit(float64(arr1D, arr1D, float64))
def interp1d_linear(x, p, x0):
    '''
    1D linear interpolation
       x 1D array. This must be monotocically increasing. Note the routine does not check this.
       p 1D array
       x0 point to interpolate
    '''

    size = len(x)
    if x0 < x[0]:
        return p[0]
    if x0 > x[size-1]:
        return p[size-1]

    i = np.argmin(np.abs(x - x0))
    if i == 0:
        pass
    elif i == size - 1:
        i = size - 2
    else:
        if x[i] > x0:
            i = i-1
        else:
            i = i

    assert x[i+1]-x[i] > 0, "input array has to be monotonic"

    d1 = p[i+1] * (x0 - x[i])/(x[i+1]-x[i]) + \
        p[i] * (x[i+1] - x0)/(x[i+1]-x[i])
    return d1


@njit(float64(arr1D, arr1D, float64))
def interp1d_cubic(x, p, x0):
    '''
    1D cubic herimit interpolation
       x 1D array. This must be monotonically increasing. Note the routine does not check this.
       p 1D array
       x0 point to interpolate
    '''
    size = len(x)
    if x0 < x[1]:
        return interp1d_linear(x, p, x0)
    if x0 > x[size-2]:
        return interp1d_linear(x, p, x0)

    i = np.argmin(np.abs(x - x0))
    if i == 1:
        pass
    elif i == size - 2:
        i = size - 3
    else:
        if x[i] > x0:
            i = i-1
        else:
            i = i

    xx = (x0 - x[i])/(x[i+1] - x[i])

    d1 = (x[i] - x[i-1])
    d2 = (x[i+1] - x[i])
    d3 = (x[i+2] - x[i+1])

    assert d1 > 0, "input array has to be monotonic"
    assert d2 > 0, "input array has to be monotonic"
    assert d3 > 0, "input array has to be monotonic"

    f0 = p[i]
    f1 = p[i+1]
    f0p = (-p[i-1]*d2 - p[i]*(d1 - d2) + p[i+1]*d1)/2/d1/d2*(x[i+1] - x[i])
    f1p = (-p[i]*d3 - p[i+1]*(d2 - d3) + p[i+2]*d2)/2/d2/d3*(x[i+1] - x[i])

    a = 2*f0 - 2*f1 + f0p + f1p
    b = -3*f0 + 3*f1 - 2*f0p-f1p
    c = f0p
    d = f0

    return a*xx**3 + b*xx**2 + c*xx**1 + d

    '''
    return p[i] + 0.5 * x*(p[i+1] - p[i-1] + x*(2.0*p[i-1] - 5.0*p[i] + 4.0*p[i+1] - p[i+2] + x*(3.0*(p[i] - p[i+1]) + p[i+2] - p[i-1])))
    '''


@njit(float64(arr1D, arr1D, arr2D, float64, float64))
def interp2d_nearest(x, y, p, x0, y0):
    '''
    2D nearest neighbor interpolation
       x 1D array. This must be monotocically increasing. Note the routine does not check this.
       y 1D array. This must be monotocically increasing. Note the routine does not check this.
       p 2D array
       x0 point to interpolate
       y0 point to interpolate
    '''

    i = np.argmin(np.abs(x - x0))
    j = np.argmin(np.abs(y - y0))
    return p[j, i]


@njit(float64(arr1D, arr1D, arr2D, float64, float64))
def interp2d_linear(x, y, p, x0, y0):
    '''
    2D linear interpolation
       x 1D array. This must be monotocically increasing. Note the routine does not check this.
       y 1D array. This must be monotocically increasing. Note the routine does not check this.
       p 2D array
       x0 point to interpolate
       y0 point to interpolate
    '''

    size = len(y)
    if y0 < y[0]:
        return interp1d_linear(x, p[0, :], x0)
    if y0 > y[-1]:
        return interp1d_linear(x, p[-1, :], x0)

    i = np.argmin(np.abs(y - y0))
    if i == 0:
        pass
    elif i == size - 1:
        i = size - 2
    else:
        if y[i] > y0:
            i = i-1
        else:
            i = i

    a0 = interp1d_linear(x, p[i, :], x0)
    a1 = interp1d_linear(x, p[i+1, :], x0)

    assert y[i+1]-y[i] > 0, "input array has to be monotonic"

    return a0 * (y[i+1] - y0)/(y[i+1] - y[i]) + a1 * (y0 - y[i])/(y[i+1]-y[i])


@njit(float64(arr1D, arr1D, arr2D, float64, float64))
def interp2d_cubic(x, y, p, x0, y0):
    '''
    2D cubic interpolation
       x 1D array. This must be monotoci. Note the routine does not check this.
       y 1D array. This must be monotoci. Note the routine does not check this.
       p 2D array
       x0 point to interpolate
       y0 point to interpolate
    '''

    size = len(y)
    if y0 < y[1]:
        a0 = interp1d_cubic(x, p[0, :], x0)
        a1 = interp1d_cubic(x, p[1, :], x0)
        return a0 * (y[1] - y0)/(y[1] - y[0]) + a1 * (y0 - y[0])/(y[1]-y[0])
    if y0 > y[-2]:
        a0 = interp1d_cubic(x, p[-2, :], x0)
        a1 = interp1d_cubic(x, p[-1, :], x0)
        return a0 * (y[-1] - y0)/(y[-1] - y[-2]) + a1 * (y0 - y[-2])/(y[-1]-y[-2])

    i = np.argmin(np.abs(y - y0))
    if i == 1:
        pass
    elif i == size - 2:
        i = size - 3
    else:
        if y[i] > y0:
            i = i-1
        else:
            i = i

    a0 = interp1d_cubic(x, p[i-1, :], x0)
    a1 = interp1d_cubic(x, p[i, :], x0)
    a2 = interp1d_cubic(x, p[i+1, :], x0)
    a3 = interp1d_cubic(x, p[i+2, :], x0)

    yy = (y0 - y[i])/(y[i+1] - y[i])

    d1 = (y[i] - y[i-1])
    d2 = (y[i+1] - y[i])
    d3 = (y[i+2] - y[i+1])

    assert d1 > 0, "input array has to be monotonic"
    assert d2 > 0, "input array has to be monotonic"
    assert d3 > 0, "input array has to be monotonic"

    f0 = a1
    f1 = a2
    f0p = (-a0*d2 - a1*(d1 - d2) + a2*d1)/2/d1/d2*(y[i+1] - y[i])
    f1p = (-a1*d3 - a2*(d2 - d3) + a3*d2)/2/d2/d3*(y[i+1] - y[i])

    a = 2*f0 - 2*f1 + f0p + f1p
    b = -3*f0 + 3*f1 - 2*f0p-f1p
    c = f0p
    d = f0

    return a*yy**3 + b*yy**2 + c*yy**1 + d

    '''
    return a1 + 0.5 * yy*(a2 - a0 + yy*(2.0*a0 - 5.0*a1 + 4.0*a2 - a3 + yy*(3.0*(a1 - a2) + a3 - a0)))
    '''


@njit(float64(arr1D, arr1D, arr1D, arr3D, float64, float64, float64))
def interp3d_nearest(x, y, z, p, x0, y0,  z0):
    '''
    3D nearest neighbor interpolation
       x 1D array. This must be monotocically increasing. Note the routine does not check this.
       y 1D array. This must be monotocically increasing. Note the routine does not check this.
       z 1D array. This must be monotocically increasing. Note the routine does not check this.
       p 3D array
       x0 point to interpolate
       y0 point to interpolate
       z0 point to interpolate
    '''

    i = np.argmin(np.abs(x - x0))
    j = np.argmin(np.abs(y - y0))
    k = np.argmin(np.abs(z - z0))
    return p[k, j, i]


@njit(float64(arr1D, arr1D, arr1D, arr3D, float64, float64, float64))
def interp3d_linear(x, y, z, p, x0, y0,  z0):
    '''
    3D linear interpolation
       x 1D array. This must be monotocically increasing. Note the routine does not check this.
       y 1D array. This must be monotocically increasing. Note the routine does not check this.
       z 1D array. This must be monotocically increasing. Note the routine does not check this.
       p 3D array
       x0 point to interpolate
       y0 point to interpolate
       z0 point to interpolate
    '''

    if z0 < z[0]:
        return interp2d_linear(x, y, p[0, :, :], x0, y0)
    if z0 > z[-1]:
        return interp2d_linear(x, y, p[-1, :, :], x0, y0)

    size3 = len(z)
    k = np.argmin(np.abs(z - z0))
    if k == 0:
        pass
    elif k == size3 - 1:
        k = size3 - 2
    else:
        if z[k] > z0:
            k = k-1
        else:
            k = k

    a0 = interp2d_linear(x, y, p[k, :, :], x0, y0)
    a1 = interp2d_linear(x, y, p[k+1, :, :], x0, y0)

    assert z[k+1]-z[k] > 0, "input array has to be monotonic"

    return a0 * (z[k+1] - z0)/(z[k+1] - z[k]) + a1 * (z0 - z[k])/(z[k+1]-z[k])


@njit(float64(arr1D, arr1D, arr1D, arr3D, float64, float64, float64))
def interp3d_cubic(x, y, z, p, x0, y0,  z0):
    '''
    3D cubic interpolation
       x 1D array. This must be monotoci. Note the routine does not check this.
       y 1D array. This must be monotoci. Note the routine does not check this.
       z 1D array. This must be monotoci. Note the routine does not check this.
       p 3D array
       x0 point to interpolate
       y0 point to interpolate
       z0 point to interpolate
    '''

    if z0 < z[1]:
        a0 = interp2d_cubic(x, y, p[0, :, :], x0, y0)
        a1 = interp2d_cubic(x, y, p[1, :, :], x0, y0)
        return a0 * (z[1] - z0)/(z[1] - z[0]) + a1 * (z0 - z[0])/(z[1]-z[0])
    if z0 > z[-2]:
        a0 = interp2d_cubic(x, y, p[-2, :, :], x0, y0)
        a1 = interp2d_cubic(x, y, p[-1, :, :], x0, y0)
        return a0 * (z[-1] - z0)/(z[-1] - z[-2]) + a1 * (z0 - z[-2])/(z[-1]-z[-2])

    size3 = len(z)
    k = np.argmin(np.abs(z - z0))
    if k == 1:
        pass
    elif k == size3 - 2:
        i = size3 - 3
    else:
        if z[k] > z0:
            k = k-1
        else:
            k = k

    a0 = interp2d_cubic(x, y, p[k-1, :, :], x0, y0)
    a1 = interp2d_cubic(x, y, p[k, :, :], x0, y0)
    a2 = interp2d_cubic(x, y, p[k+1, :, :], x0, y0)
    a3 = interp2d_cubic(x, y, p[k+2, :, :], x0, y0)

    zz = (z0 - z[k])/(z[k+1] - z[k])

    d1 = (z[k] - z[k-1])
    d2 = (z[k+1] - z[k])
    d3 = (z[k+2] - z[k+1])

    assert d1 > 0, "input array has to be monotonic"
    assert d2 > 0, "input array has to be monotonic"
    assert d3 > 0, "input array has to be monotonic"

    f0 = a1
    f1 = a2
    f0p = (-a0*d2 - a1*(d1 - d2) + a2*d1)/2/d1/d2*(z[k+1] - z[k])
    f1p = (-a1*d3 - a2*(d2 - d3) + a3*d2)/2/d2/d3*(z[k+1] - z[k])

    a = 2*f0 - 2*f1 + f0p + f1p
    b = -3*f0 + 3*f1 - 2*f0p-f1p
    c = f0p
    d = f0

    return a*zz**3 + b*zz**2 + c*zz**1 + d


@njit
def test1d():
    x = np.linspace(0, np.pi*6, 10)
    y = np.sin(x/10*np.pi)

    new_x = np.linspace(0, np.pi*6, 1000)
    #new_y = np.zeros(len(new_x), dtype=float64)
    # for i in range(len(new_x)):
    #    new_y[i] = interp1d_nearest(x, y, new_x[i])
    #

    new_y0 = np.array([interp1d_nearest(x, y, xx) for xx in new_x])

    new_y1 = np.array([interp1d_linear(x, y, xx) for xx in new_x])

    new_y3 = np.array([interp1d_cubic(x, y, xx) for xx in new_x])

    xo = np.linspace(0, np.pi*6, 1000)
    yo = np.sin(xo/10*np.pi)
    return x, y, xo, yo, new_x, new_y0, new_y1, new_y3


@njit
def test2d():
    size1 = 15
    size2 = 21

    x = np.linspace(-3, 3, size1)
    y = np.linspace(-3, 3, size2)
    p = np.zeros((len(y), len(x)))
    for j in range(len(y)):
        for i in range(len(x)):
            p[j, i] = np.exp(-(x[i]**2 + y[j]**2)/3)

    x2 = np.hstack((np.linspace(-3, 1, size1*3)
                   [:-1], np.linspace(1, 3, size1*1000)))
    y2 = np.hstack((np.linspace(-3, 1, size2*3)
                   [:-1], np.linspace(1, 3, size2*1000)))
    p2 = np.zeros((len(y2), len(x2)))
    for j in range(len(y2)):
        for i in range(len(x2)):
            p2[j, i] = np.exp(-(x2[i]**2 + y2[j]**2)/3)

    new_x = np.linspace(-3, 3, 2500)
    new_y = np.linspace(-3, 3, 30)

    shape = (len(new_y), len(new_x))
    out0 = np.zeros(shape)
    out1 = np.zeros(shape)
    out2 = np.zeros(shape)

    for j in range(shape[0]):
        for i in range(shape[1]):
            out0[j, i] = interp2d_nearest(x2, y2, p2, new_x[i], new_y[j])
            out1[j, i] = interp2d_linear(x2, y2, p2, new_x[i], new_y[j])
            out2[j, i] = interp2d_cubic(x2, y2, p2, new_x[i], new_y[j])

    return x, y, p, new_x, new_y, out0, out1, out2


@njit
def test3d():
    size1, size2, size3 = 9, 7, 11

    x = np.linspace(-3, 3, size1)
    y = np.linspace(-3, 3, size2)
    z = np.linspace(-3, 3, size3)

    p = np.zeros((size3, size2, size1))
    for k in range(size3):
        for j in range(size2):
            for i in range(size1):
                p[k, j, i] = np.exp(-(x[i]**2 + y[j]**2 + z[k]**2)/3)

    new_x = np.linspace(-3, 3, 25)
    new_y = np.linspace(-3, 3, 15)
    new_z = np.linspace(-3, 3, 31)

    shape = (len(new_z), len(new_y), len(new_x))
    out0 = np.zeros(shape)
    out1 = np.zeros(shape)
    out2 = np.zeros(shape)

    for k in range(shape[0]):
        for j in range(shape[1]):
            for i in range(shape[2]):
                out0[k, j, i] = interp3d_nearest(
                    x, y, z, p, new_x[i], new_y[j], new_z[k])
                out1[k, j, i] = interp3d_linear(
                    x, y, z, p, new_x[i], new_y[j], new_z[k])
                out2[k, j, i] = interp3d_cubic(
                    x, y, z, p, new_x[i], new_y[j], new_z[k])

    return x, y, z, p, new_x, new_y, new_z, out0, out1, out2


print()



'''
   utility routines
'''
import os
import numpy as np
import resource
from petram.mfem_config import use_parallel
import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('Utils')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem


def file_write(fid, *args):
    txt = ' '.join([str(x) for x in args])
    print(txt)
    fid.write(txt + "\n")


def print_mem(myid=0):
    mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    print("memory usage (" + str(myid) + "): " + str(mem))


def set_array_attribute(v, base, suffix, values):
    for s, vv in zip(suffix, values):
        v[base + s] = vv
    return v


def txt2indexlist(txt):
    try:
        return [int(x) for x in txt.split(',')]
    except:
        raise ValueError("can not convert text to index list")


def eval_expression(expr, mesh, battr, ind_vars, ns, use_dom=False):
    '''
    example:
        expr = 'x', or 'sin(x)'
        ibdry = 3 (boundary attribute number)
        ind_vars ['x', 'y', 'z']
    '''
    def _do_eval(code, verts):
        l = {n: verts[k] for k, n in enumerate(ind_vars)}
        return eval(code, ns, l)

    if use_dom:
        get_array = mesh.GetDomainArray
        get_element = mesh.GetElement
    else:
        get_array = mesh.GetBdrArray
        get_element = mesh.GetBdrElement

    ibdr = get_array(battr)

    code = compile(expr, '<string>', 'eval')

    iverts = np.stack([get_element(i).GetVerticesArray() for i in ibdr])
    locs = np.stack([np.stack([mesh.GetVertexArray(k) for k in ivert])
                     for ivert in iverts])
    data = np.stack([np.stack([_do_eval(code, mesh.GetVertexArray(k))
                               for k in ivert])
                     for ivert in iverts])
    return locs, data


def eval_expr(model, engine, expr, battrs, phys=None):
    '''
    expr = expression to evaluate
    battrs = list of boundary attribute
    '''
    from petram.model import Bdry

    if phys is None:
        phys = model['Phys'][list(model['Phys'])[0]]
    else:
        phys = model['Phys'][phys]

    ret = {}
    mesh = engine.get_mesh()
    engine.assign_sel_index(phys)

    use_dom = (phys.dim == 2 or phys.dim == 1)
    ns = phys._global_ns
    battrs = list(np.atleast_1d(eval(battrs, ns, {})).flatten().astype(int))

    for m in phys.walk():
        for battr in battrs:
            ind_vars = m.get_independent_variables()
            ret[battr] = eval_expression(
                expr, mesh, battr, ind_vars, ns, use_dom=use_dom)
    return ret


def eval_sol(sol, battrs, dim=0):
    mesh = sol.FESpace().GetMesh()
    ibdr = np.hstack([mesh.GetBdrArray(battr) for battr in battrs])
    if len(ibdr) == 0:
        return None, None
    iverts = np.stack([mesh.GetBdrElement(i).GetVerticesArray() for i in ibdr])
    locs = np.stack([np.stack([mesh.GetVertexArray(k) for k in ivert])
                     for ivert in iverts])

    data = sol.GetNodalValues(dim)
    data = data[iverts.flatten()].reshape(iverts.shape)

    return locs, data


def eval_loc(sol, battrs):
    mesh = sol.FESpace().GetMesh()
    ibdr = np.hstack([mesh.GetBdrArray(battr) for battr in battrs])
    if len(ibdr) == 0:
        return None, None
    iverts = np.stack([mesh.GetBdrElement(i).GetVerticesArray() for i in ibdr])
    locs = np.stack([np.stack([mesh.GetVertexArray(k) for k in ivert])
                     for ivert in iverts])
    return locs


def get_pkg_datafile(pkg, *path):
    '''
    return package data

    ex) get_pkg_datafile(petram.geom, 'icon', 'line.png')
    '''
    file = getattr(pkg, '__file__')
    root = os.path.abspath(os.path.dirname(file))
    return os.path.join(os.path.dirname(root), 'data', *path)


def get_evn_petram_root():
    petram = os.getenv("PetraM")
    return petram


def get_evn_twopiroot():
    petram = os.getenv("TwoPiRoot")
    return petram


def check_cluster_access():
    petram = get_evn_twopiroot()
    return os.path.exists(os.path.join(petram, "etc", "cluster_access"))


def check_addon_access():
    petram = get_evn_twopiroot()
    if os.path.exists(os.path.join(petram, "etc", "addon_access")):
        return "any"
    return "none"


''''
 pv (paired value) handling (used in processing GUI data)
'''


def pv_get_gui_value(mm, paired_var):
    '''
    return value for GUI panel
    '''
    mfem_physroot = mm.get_root_phys().parent
    paired_var = mm.paired_var
    if paired_var is not None:
        try:
            var_s = mfem_physroot[paired_var[0]].dep_vars
            name1 = var_s[paired_var[1]]
            model1 = paired_var[0]
        except BaseException:
            paired_var = None

    if paired_var is None:
        names = mm.get_root_phys().dep_vars
        if len(names) > 0:
            name1 = names[0]
        else:
            name1 = ""
        model1 = mm.get_root_phys().name()
    var = name1 + " ("+model1 + ")"
    return var, paired_var


def pv_from_gui_value(mm, value):
    '''
    return paired_var from GUI input
    '''
    mfem_physroot = mm.get_root_phys().parent
    names, pnames, pindex = mfem_physroot.dependent_values(include_disabled=True)

    if len(value) == 0:
        # v[0] could be '' if object is based to a tree.
        idx = 0
    else:
        tmp = str(value).split("(")[0].strip()
        if tmp in names:
            idx = names.index(tmp)
        else:
            idx = 0

    if len(pnames) == 0:
        return None
    paired_var = (pnames[idx], pindex[idx])
    return paired_var


def pv_panel_param(mm, label):
    '''
    panel value for paired_var
    '''
    from wx import CB_READONLY
    mfem_physroot = mm.get_root_phys().parent
    names, pnames, _pindex = mfem_physroot.dependent_values(include_disabled=True)
    names = [n+" ("+p + ")" for n, p in zip(names, pnames)]

    ll1 = [label, "S", 4,
           {"style": CB_READONLY, "choices": names}]
    return ll1


''''
 pm (paired model) handling (used in processing GUI data)
'''


def pm_get_gui_value(mm, paired_model):
    '''
    return value for GUI panel
    '''
    mfem_physroot = mm.get_root_phys().parent

    paired_model = mm.paired_model
    if paired_model not in mfem_physroot:
        model1 = mm.get_root_phys().name()
        paired_model = None
    else:
        model1 = paired_model
    return model1, paired_model


def pm_from_gui_value(mm, value):
    '''
    return paired_model from GUI input
    '''
    mfem_physroot = mm.get_root_phys().parent
    names = [x.name() for x in mfem_physroot.get_children()]

    if len(names) == 0:
        return None

    if len(value) == 0:
        # v[0] could be '' if object is based to a tree.
        idx = 0
    else:
        if value in names:
            idx = names.index(value)
        else:
            idx = 0

    paired_model = names[idx]
    return paired_model


def pm_panel_param(mm, label):
    '''
    panel value for paired_var
    '''
    from wx import CB_READONLY
    mfem_physroot = mm.get_root_phys().parent
    names = [x.name() for x in mfem_physroot.get_children()]

    ll1 = [label, "S", 4,
           {"style": CB_READONLY, "choices": names}]
    return ll1



import numpy as np
import os

from petram.mfem_config import use_parallel
if use_parallel:
    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    smyid = '.'+'{:0>6d}'.format(myid)
else:
    myid = 0
    smyid = ''


def list_probes(dir):
    od = os.getcwd()
    os.chdir(dir)
    filenames = [name for name in os.listdir() if name.startswith('probe_')]

    if use_parallel:
        filenames = [f for f in filenames if '.'+f.split('.')[-1] == smyid]

    probenames = [n[6:] for n in filenames]
    os.chdir(od)

    return filenames, probenames


def load_probe(name):
    fid = open(name, 'r')
    format = int(fid.readline().split(':')[-1])

    if format == 0:
        value = load_format_0(fid)
        ydata = np.squeeze(value[:, 1:].transpose())
        xdata = value[:, 0]
        xdata = {'time': xdata}
    if format == 1:
        xdata, ydata, names = load_format_1(fid)
        xdata = {n: xdata[k] for k, n in enumerate(names)}

    fid.close()

    return xdata, ydata


def load_probes(dir, names):
    data = [load_probe(os.path.join(dir, n)) for n in names]

    xdata = data[0][0]
    ydata = [x[1] for x in data]

    #print("ydata shape", [x[1].shape for x in data])
    ydata = np.vstack(ydata)
    ydata = np.squeeze(ydata)

    return xdata, ydata


def load_format_0(fid):
    lines = fid.readlines()
    data = [[float(x) for x in l.split(',') if x.strip() != ''] for l in lines]
    data = np.array(data)
    return data


def load_format_1(fid):
    from petram.helper.variables import Variable, var_g

    xsize = int(fid.readline())
    xnames = [x.strip() for x in fid.readline().split(',')]

    lines = fid.readlines()

    data = [eval(l, var_g, {}) for l in lines]

    xdata = np.array([x[:xsize] for x in data]).transpose()
    ydata = np.array([x[xsize:] for x in data]).transpose()

    return xdata, ydata, xnames


class Probe(object):
    def __new__(cls, *args, **kargs):
        root_only = kargs.pop("root_only", False)
        if myid == 0 or not root_only:
            return object.__new__(cls)
        else:
            return None

    def __init__(self, name, idx=-1, xnames=None, **kwargs):
        '''
        idx is idx in blockvector. could be -1 in such case, Probe can not load data from
        sol. This is used in Parametric to gather all probe later
        '''
        self.name = name
        self.xnames = ['time'] if xnames is None else xnames
        self.sig = []
        self.t = []
        self.idx = idx
        self.finalized = False

    def write_file(self, filename=None, format=1):
        valid = self.finalize()
        if not valid:
            return

        if filename is None:
            filename = 'probe_'+self.name + smyid

        fid = open(filename, 'w')

        if format == 0:
            fid.write("format : 0\n")
        else:
            fid.write("format : 1\n")
            fid.write(str(len(self.time[0]))+"\n")
            fid.write(','.join(self.xnames)+"\n")

        for x, t in zip(self.sig_f, self.time):
            txt1 = ', '.join([str(xx) for xx in t])
            txt2 = ', '.join([str(xx) for xx in x])
            fid.write(txt1 + ', ' + txt2 + "\n")
        fid.close()

    def append_sol(self, sol, t=0.0):
        self.sig.append(np.atleast_1d(sol[self.idx].toarray().flatten()))
        self.t.append(np.atleast_1d(t))

    def append_value(self, value, t=0.0):
        self.sig.append(np.atleast_1d(value))
        self.t.append(np.atleast_1d(t))

    def current_value(self, sol):
        return np.atleast_1d(sol[self.idx].toarray().flatten())

    def print_signal(self):
        if not self.finalized:
            self.finalize()

    def finalize(self):
        if len(self.sig) == 0:
            self.sig_f = -1
            self.valid = False
        else:
            shapes = [x.shape for x in self.sig]

            if np.all([len(x.shape) == 1 for x in self.sig]):
                self.sig_f = np.vstack(self.sig)
            else:
                shape2 = max([np.prod(x.shape) for x in self.sig])
                iscomplex = np.any([np.iscomplexobj(x) for x in self.sig])
                if iscomplex:
                    data = np.full((len(self.sig), shape2), np.nan, dtype=self.sig[0].dtype)
                else:
                    data = np.full((len(self.sig), shape2), np.nan, dtype=self.sig[0].dtype)
                for k, x in enumerate(self.sig):
                    l = np.prod(x.shape)
                    data[k, :l] = x.flatten()

                self.sig_f = data

            self.time = self.t
            self.valid = self.sig_f.size != 0
            #self.valid = True
        self.finalized = True
        return self.valid



"""

  Petra-M setuptools based setup module.

"""
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

setup(
    name='PetraM_Base',

    version='2.1.26',

    description='PetraM base package',
    long_description=long_description,
    url='https://github.com/piScope/PetraM',
    author='S. Shiraiwa',
    author_email='shiraiwa@prenceton.edu',
    license='GNUv3',

    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='MFEM physics',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=[],
    extras_require={},
    # , '':['data/*.mesh']},
    package_data={'petram': ['data/icon/*.png', 'data/*Ops']},
    data_files=[],
    entry_points={},
)



'''
    Run('xxxx.nas'): generate xxxx.mesh
'''
import os 
from petram.mesh.nastran2mfem import NASReader, write_nas2mfem
from os.path import expanduser


def nas2mfem(file = None, offset=None):
    if file is None:
        from ifigure.widgets.dialog import read
        path = read(message='Select NAS file to read', wildcard='*.nas')
        if path == '': return
    else:
        path = expanduser(file)
    if offset is None: offset = [0.0, 0.0, 0.0]
    mname = os.path.basename(path)
    dirname = os.path.dirname(path)
    if '.' in mname: 
        mname = '.'.join(mname.split('.')[:-1])
    mname = mname + '.mesh'
    print('reading file '+ path)
    reader = NASReader(path)
    reader.load()
    write_nas2mfem(os.path.join(dirname, mname),  reader, offset=offset)

ans(nas2mfem(*args, **kwargs))



def start_engine(mfemmodel=None):
    if mfemmodel is None: 
        mfemmodel = param.eval('mfem_model')
    if not 'engine' in globals():
        from petram.engine import SerialEngine
        engine = SerialEngine(model=mfemmodel)
        model.variables.setvar('engine', engine)
    else:
        engine = model.variables.getvar('engine')
    return engine
ans(start_engine())



from petram.solver.mumps_model import MUMPSPreconditioner
from petram.mfem_config import use_parallel
import numpy as np

from petram.debug import flush_stdout
from petram.namespace_mixin import NS_mixin
from .solver_model import LinearSolverModel, LinearSolver

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('IterativeSolverModel')

if use_parallel:
    from petram.helper.mpi_recipes import *
    from mfem.common.parcsr_extra import *
    import mfem.par as mfem
    default_kind = 'hypre'

    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    smyid = '{:0>6d}'.format(myid)
    from mfem.common.mpi_debug import nicePrint

else:
    import mfem.ser as mfem
    default_kind = 'scipy'

SparseSmootherCls = {"Jacobi": (mfem.DSmoother, 0),
                     "l1Jacobi": (mfem.DSmoother, 1),
                     "lumpedJacobi": (mfem.DSmoother, 2),
                     "GS": (mfem.GSSmoother, 0),
                     "forwardGS": (mfem.GSSmoother, 1),
                     "backwardGS": (mfem.GSSmoother, 2),
                     "MUMPS": (MUMPSPreconditioner, None), }

choices_a = ['CG', 'GMRES', 'FGMRES', 'BiCGSTAB',
             'MINRES', 'SLI', 'Nested FGMRES']
choices_b = ['CG', 'GMRES', 'FGMRES', 'BiCGSTAB', 'MINRES', 'MUMPS']

single1_elp = [["log_level", -1, 400, {}],
               ["max  iter.", 200, 400, {}],
               ["rel. tol", 1e-7, 300, {}],
               ["abs. tol.", 1e-7, 300, {}], ]
single2_elp = [["log_level", -1, 400, {}],
               ["max  iter.", 200, 400, {}],
               ["rel. tol", 1e-7, 300, {}],
               ["abs. tol.", 1e-7, 300, {}],
               ["restart(kdim)", 50, 400, {}]]

nsingle1_elp = [["log_level", -1, 400, {}],
                ["max  iter.", 200, 400, {}], ]
nsingle2_elp = [["log_level", -1, 400, {}],
                ["max  iter.", 200, 400, {}],
                ["restart(kdim)", 50, 400, {}], ]
nmumps_elp = [['MUMPS Config', '', 0, {}], ]
nested_elp = [["log_level", -1, 400, {}],
              ["max  iter.", 200, 400, {}],
              ["rel. tol", 1e-7, 300, {}],
              ["abs. tol.", 1e-7, 300, {}],
              ["restart(kdim)", 50, 400, {}],
              [None, None, 34, [{'text': "Inner Solver",
                                 'choices': choices_b},
                                {'elp': nsingle1_elp},  # CG
                                {'elp': nsingle2_elp},  # GMRES
                                {'elp': nsingle2_elp},  # FGMRES
                                {'elp': nsingle1_elp},  # BiCGSTAB
                                {'elp': nsingle1_elp},  # MINRES
                                {'elp': nmumps_elp}, ]]  # MUMPS
              , ]


class Iterative(LinearSolverModel, NS_mixin):
    hide_ns_menu = True
    has_2nd_panel = False
    accept_complex = False
    always_new_panel = False

    def __init__(self, *args, **kwargs):
        LinearSolverModel.__init__(self, *args, **kwargs)
        NS_mixin.__init__(self, *args, **kwargs)

    def init_solver(self):
        pass

    def panel1_param(self):
        import wx
        from petram.pi.widget_smoother import WidgetSmoother

        smp1 = [None, None, 99, {"UI": WidgetSmoother, "span": (1, 2)}]

        mm = [[None, self.use_block_symmetric, 3,
               {"text": "block symmetric format"}], ]

        return [[None, None, 34, [{'text': "Solver", 'choices': choices_a},
                                  {'elp': single1_elp},  # CG
                                  {'elp': single2_elp},  # GMRES
                                  {'elp': single2_elp},  # FGMRES
                                  {'elp': single1_elp},  # BiCGSTAB
                                  {'elp': single1_elp},  # MINRES
                                  {'elp': single1_elp},  # SLI
                                  {'elp': nested_elp},  # Nested FGMRES
                                  ], ],
                [None, [False, [''], [[], ]], 27, [{'text': 'advanced mode'},
                                                   {'elp': [
                                                       ['preconditioner', '', 0, None], ]},
                                                   {'elp': [smp1, ]}], ],
                [None, self.write_mat, 3, {"text": "write matrix"}],
                [None, self.assert_no_convergence, 3,
                    {"text": "check converegence"}],
                [None, self.use_ls_reducer, 3, {
                    "text": "Reduce linear system when possible"}],
                [None, (self.merge_real_imag, (self.use_block_symmetric,)),
                 27, ({"text": "Use ComplexOperator"}, {"elp": mm},)], ]

    def get_panel1_value(self):
        # this will set _mat_weight
        from petram.solver.solver_model import SolveStep
        p = self.parent
        while not isinstance(p, SolveStep):
            p = p.parent
            if p is None:
                assert False, "Solver is not under SolveStep"
        num_matrix = p.get_num_matrix(self.get_phys())

        all_dep_vars = self.root()['Phys'].all_dependent_vars(num_matrix,
                                                              self.get_phys(),
                                                              self.get_phys_range())

        prec = [x for x in self.preconditioners if x[0] in all_dep_vars]
        names = [x[0] for x in prec]
        for n in all_dep_vars:
            if not n in names:
                prec.append((n, ['None', 'None']))
        self.preconditioners = prec

        single1 = [int(self.log_level), int(self.maxiter),
                   self.reltol, self.abstol]
        single2 = [int(self.log_level), int(self.maxiter),
                   self.reltol, self.abstol, int(self.kdim)]
        single1in = [int(self.log_level_in), int(self.maxiter_in), ]
        single2in = [int(self.log_level_in), int(self.maxiter_in),
                     int(self.kdim_in)]
        mumpsin = [self.mumps_in]
        nested = [int(self.log_level), int(self.maxiter),
                  self.reltol, self.abstol, int(self.kdim),
                  [self.solver_type_in, single1in, single2in, single2in,
                   single1in, single1in, mumpsin], ]
        value = ([self.solver_type, single1, single2, single2,
                  single1, single1, single1,
                  nested],
                 (self.adv_mode, [self.adv_prc, ], [self.preconditioners, ]),
                 self.write_mat, self.assert_no_convergence,
                 self.use_ls_reducer,
                 (self.merge_real_imag, [self.use_block_symmetric, ]),)

        return value

    def import_panel1_value(self, v):
        self.solver_type = str(v[0][0])
        idx = choices_a.index(self.solver_type)
        vv = v[0][idx + 1]
        self.log_level = int(vv[0])
        self.maxiter = int(vv[1])
        self.reltol = vv[2]
        self.abstol = vv[3]

        if len(vv) > 4:
            self.kdim = int(vv[4])

        if self.solver_type.startswith('Nested'):
            self.solver_type_in = vv[5][0]
            idx = choices_b.index(self.solver_type_in)
            vv = vv[5][idx + 1]
            if self.solver_type_in != 'MUMPS':
                self.log_level_in = int(vv[0])
                self.maxiter_in = int(vv[1])
                if len(vv) > 2:
                    self.kdim_in = int(vv[2])
            else:
                self.mumps_in = vv[0]

        self.preconditioners = v[1][2][0]
        self.write_mat = bool(v[2])
        self.assert_no_convergence = bool(v[3])
        self.use_ls_reducer = bool(v[4])
        self.adv_mode = v[1][0]
        self.adv_prc = v[1][1][0]
        self.merge_real_imag = bool(v[5][0])
        self.use_block_symmetric = bool(v[5][1][0])

    def attribute_set(self, v):
        v = super(Iterative, self).attribute_set(v)
        v['solver_type'] = 'GMRES'
        v['log_level'] = 0
        v['maxiter'] = 200
        v['reltol'] = 1e-7
        v['abstol'] = 1e-7
        v['kdim'] = 50

        v['solver_type_in'] = 'GMRES'
        v['log_level_in'] = 0
        v['maxiter_in'] = 200
        v['reltol_in'] = 1e-7
        v['abstol_in'] = 1e-7
        v['kdim_in'] = 50
        v['mumps_in'] = ''

        v['printit'] = 1
        v['preconditioner'] = ''
        v['preconditioners'] = []
        v['write_mat'] = False
        v['assert_no_convergence'] = True
        v['use_ls_reducer'] = False
        v['adv_mode'] = False
        v['adv_prc'] = ''
        v['merge_real_imag'] = False
        v['use_block_symmetric'] = False
        return v

    def verify_setting(self):
        if not self.parent.assemble_real:
            for phys in self.get_phys():
                if phys.is_complex() and not self.merge_real_imag:
                    return False, "Iterative does not support complex.", "A complex problem must be converted to a real value problem"
        return True, "", ""

    def does_linearsolver_choose_linearsystem_type(self):
        return True

    def linear_system_type(self, assemble_real, phys_real):
        if phys_real:
            if assemble_real:
                dprint1("Use assemble-real is only for complex value problem !!!!")
                return 'blk_interleave'
            else:
                return 'blk_interleave'

        # below phys is complex

        # merge_real_imag -> complex operator
        if self.merge_real_imag and self.use_block_symmetric:
            return 'blk_merged_s'
        elif self.merge_real_imag and not self.use_block_symmetric:
            return 'blk_merged'
        elif assemble_real:
            return 'blk_interleave'
        else:
            assert False, "complex problem must assembled using complex operator or expand as real value problem"
        # return None

    def real_to_complex(self, solall, M):
        if self.merge_real_imag:
            return self.real_to_complex_merged(solall, M)
        else:
            return self.real_to_complex_interleaved(solall, M)

    def real_to_complex_interleaved(self, solall, M):
        if use_parallel:
            from mpi4py import MPI
            myid = MPI.COMM_WORLD.rank

            offset = M.RowOffsets().ToList()
            of = [np.sum(MPI.COMM_WORLD.allgather(np.int32(o)))
                  for o in offset]
            if myid != 0:
                return

        else:
            offset = M.RowOffsets()
            of = offset.ToList()

        rows = M.NumRowBlocks()
        s = solall.shape
        nb = rows // 2
        i = 0
        pt = 0
        result = np.zeros((s[0] // 2, s[1]), dtype='complex')
        for j in range(nb):
            l = of[i + 1] - of[i]
            result[pt:pt + l, :] = (solall[of[i]:of[i + 1], :]
                                    + 1j * solall[of[i + 1]:of[i + 2], :])
            i = i + 2
            pt = pt + l

        return result

    def real_to_complex_merged(self, solall, M):
        if use_parallel:
            from mpi4py import MPI
            myid = MPI.COMM_WORLD.rank

            offset = M.RowOffsets().ToList()
            of = [np.sum(MPI.COMM_WORLD.allgather(np.int32(o)))
                  for o in offset]
            if myid != 0:
                return

        else:
            offset = M.RowOffsets()
            of = offset.ToList()
        dprint1(of)
        rows = M.NumRowBlocks()
        s = solall.shape
        i = 0
        pt = 0
        result = np.zeros((s[0] // 2, s[1]), dtype='complex')
        for i in range(rows):
            l = of[i + 1] - of[i]
            w = int(l // 2)
            result[pt:pt + w, :] = (solall[of[i]:of[i] + w, :]
                                    + 1j * solall[(of[i] + w):of[i + 1], :])
            pt = pt + w
        return result

    def allocate_solver(self, is_complex=False, engine=None):
        solver = IterativeSolver(self, engine, int(self.maxiter),
                                 self.abstol, self.reltol, int(self.kdim))
        # solver.AllocSolver(datatype)
        return solver

    def get_possible_child(self):
        '''
        Preconditioners....
        '''
        choice = []
        try:
            from petram.solver.mumps_model import MUMPS
            choice.append(MUMPS)
        except ImportError:
            pass
        return choice


class IterativeSolver(LinearSolver):
    is_iterative = True

    def __init__(self, gui, engine, maxiter, abstol, reltol, kdim):
        self.maxiter = maxiter
        self.abstol = abstol
        self.reltol = reltol
        self.kdim = kdim
        LinearSolver.__init__(self, gui, engine)

    def SetOperator(self, opr, dist=False, name=None):
        self.Aname = name
        self.A = opr

        from petram.solver.linearsystem_reducer import LinearSystemReducer
        if use_parallel:
            if self.gui.use_ls_reducer:
                self.reducer = LinearSystemReducer(opr, name)
                self.M_reduced = self.make_preconditioner(self.reducer.A,
                                                          name=self.reducer.Aname,
                                                          parallel=True)
                solver = self.make_solver(self.reducer.A,
                                          self.M_reduced,
                                          use_mpi=True)
                self.reducer.set_solver(solver)
            else:
                self.M = self.make_preconditioner(self.A, parallel=True)
                self.solver = self.make_solver(self.A, self.M, use_mpi=True)
        else:
            if self.gui.use_ls_reducer:
                dprint1("Linear system reducer is not implemented in serial")
            self.M = self.make_preconditioner(self.A)
            self.solver = self.make_solver(self.A, self.M)

            self.reducer = None

    def Mult(self, b, x=None, case_base=0):
        if use_parallel:
            return self.solve_parallel(self.A, b, x)
        else:
            return self.solve_serial(self.A, b, x)

    def make_solver(self, A, M, use_mpi=False):
        maxiter = int(self.maxiter)
        atol = self.abstol
        rtol = self.reltol
        kdim = int(self.kdim)
        printit = 1

        args = (MPI.COMM_WORLD,) if use_mpi else ()

        if self.gui.solver_type.startswith("Nested"):
            solver_type = self.gui.solver_type.split(" ")[-1]
            nested = True
        else:
            solver_type = self.gui.solver_type
            nested = False
        cls = getattr(mfem, solver_type + 'Solver')

        solver = cls(*args)
        if solver_type in ['GMRES', 'FGMRES']:
            solver.SetKDim(kdim)

        if nested:
            inner_solver_type = self.gui.solver_type_in
            if inner_solver_type != "MUMPS":
                cls = getattr(mfem, inner_solver_type + 'Solver')
                inner_solver = cls(*args)
                inner_solver.SetAbsTol(0.0)
                inner_solver.SetRelTol(0.0)
                inner_solver.SetMaxIter(self.gui.maxiter_in)
                inner_solver.SetPrintLevel(self.gui.log_level_in)
                if inner_solver_type in ['GMRES', 'FGMRES']:
                    inner_solver.SetKDim(int(self.gui.kdim_in))
                inner_solver.iterative_mode = False
                inner_solver.SetOperator(A)
                inner_solver.SetPreconditioner(M)
                # return inner_solver
                prc = inner_solver
            else:
                from petram.solver.mumps_model import MUMPSBlockPreconditioner
                prc = MUMPSBlockPreconditioner(A, gui=self.gui[self.gui.mumps_in],
                                               engine=self.engine)

        else:
            prc = M
        solver._prc = prc
        solver.SetPreconditioner(prc)
        solver.SetOperator(A)

        solver.SetAbsTol(atol)
        solver.SetRelTol(rtol)
        solver.SetMaxIter(maxiter)
        solver.SetPrintLevel(self.gui.log_level)

        return solver

    def make_preconditioner(self, A, name=None, parallel=False):
        name = self.Aname if name is None else name

        if self.gui.adv_mode:
            expr = self.gui.adv_prc
            gen = eval(expr, self.gui._global_ns)
            gen.set_param(A, name, self.engine, self.gui)
            M = gen()

        else:
            prcs_gui = dict(self.gui.preconditioners)
            #assert not self.gui.parent.is_complex(), "can not solve complex"
            if self.gui.parent.is_converted_from_complex() and not self.gui.merge_real_imag:
                name = sum([[n, n] for n in name], [])

            import petram.helper.preconditioners as prcs

            g = prcs.DiagonalPrcGen(
                opr=A, engine=self.engine, gui=self.gui, name=name)
            M = g()

            pc_block = {}

            for k, n in enumerate(name):
                prctxt = prcs_gui[n][1] if parallel else prcs_gui[n][0]
                if prctxt == "None":
                    continue
                if prctxt.find("(") == -1:
                    prctxt = prctxt + "()"
                prcargs = "(".join(prctxt.split("(")[-1:])

                nn = prctxt.split("(")[0]

                if not n in pc_block:
                    # make a new one
                    dprint1(nn)
                    try:
                        blkgen = getattr(prcs, nn)
                    except BaseException:
                        if nn in self.gui._global_ns:
                            blkgen = self.gui._global_ns[nn]
                        else:
                            raise

                    blkgen.set_param(g, n)
                    blk = eval("blkgen(" + prcargs)

                    M.SetDiagonalBlock(k, blk)
                    pc_block[n] = blk
                else:
                    M.SetDiagonalBlock(k, pc_block[n])
        return M

    def write_mat(self, A, b, x, suffix=""):
        def get_block(Op, i, j):
            try:
                return Op._linked_op[(i, j)]
            except KeyError:
                return None

        offset = A.RowOffsets()
        rows = A.NumRowBlocks()
        cols = A.NumColBlocks()

        for i in range(cols):
            for j in range(rows):
                m = get_block(A, i, j)
                if m is None:
                    continue
                m.Print('matrix_' + str(i) + '_' + str(j))
        for i, bb in enumerate(b):
            for j in range(rows):
                v = bb.GetBlock(j)
                v.Print('rhs_' + str(i) + '_' + str(j) + suffix)
                #np.save('rhs_' + str(i) + '_' + str(j) + suffix, v.GetDataArray())
        if x is not None:
            for j in range(rows):
                xx = x.GetBlock(j)
                xx.Print('x_' + str(i) + '_' + str(j) + suffix)

    @flush_stdout
    def call_mult(self, solver, bb, xx):
        #print(np.sum(bb.GetDataArray()), np.sum(xx.GetDataArray()))
        solver.Mult(bb, xx)
        max_iter = solver.GetNumIterations()
        tol = solver.GetFinalNorm()

        dprint1("convergence check (max_iter, tol) ", max_iter, " ", tol)
        if self.gui.assert_no_convergence:
            if not solver.GetConverged():
                self.gui.set_solve_error(
                    (True, "No Convergence: " + self.gui.name()))
                assert False, "No convergence"

    def solve_parallel(self, A, b, x=None):
        if self.gui.write_mat:
            self. write_mat(A, b, x, "." + smyid)

        sol = []

        # solve the problem and gather solution to head node...
        # may not be the best approach

        from petram.helper.mpi_recipes import gather_vector
        offset = A.RowOffsets()
        for bb in b:
            rows = MPI.COMM_WORLD.allgather(np.int32(bb.Size()))
            #rowstarts = np.hstack((0, np.cumsum(rows)))
            dprint1("row offset", offset.ToList())
            if x is None:
                xx = mfem.BlockVector(offset)
                xx.Assign(0.0)
            else:
                xx = x

            if self.gui.use_ls_reducer:
                try:
                    self.reducer.Mult(bb, xx, self.gui.assert_no_convergence)
                except debug.ConvergenceError:
                    self.gui.set_solve_error(
                        (True, "No Convergence: " + self.gui.name()))
                    assert False, "No convergence"
            else:
                self.call_mult(self.solver, bb, xx)

            s = []
            for i in range(offset.Size() - 1):
                v = xx.GetBlock(i).GetDataArray()
                if self.gui.merge_real_imag:
                    w = int(len(v) // 2)
                    vv1 = gather_vector(v[:w])
                    vv2 = gather_vector(v[w:])
                    vv = np.hstack((vv1, vv2))
                else:
                    vv = gather_vector(v)
                if myid == 0:
                    s.append(vv)
                else:
                    pass
            if myid == 0:
                sol.append(np.hstack(s))

        if myid == 0:
            sol = np.transpose(np.vstack(sol))
            return sol
        else:
            return None

    def solve_serial(self, A, b, x=None):
        if self.gui.write_mat:
            self. write_mat(A, b, x)

        #M = self.M
        solver = self.solver

        sol = []

        for bb in b:
            if x is None:
                xx = mfem.Vector(bb.Size())
                xx.Assign(0.0)
            else:
                xx = x
                # for j in range(cols):
                #   print x.GetBlock(j).Size()
                #   print x.GetBlock(j).GetDataArray()
                #assert False, "must implement this"
            self.call_mult(solver, bb, xx)

            sol.append(xx.GetDataArray().copy())
        sol = np.transpose(np.vstack(sol))
        return sol

        '''
        prcs = dict(self.gui.preconditioners)
        name = self.Aname
        assert not self.gui.parent.is_complex(), "can not solve complex"
        if self.gui.parent.is_converted_from_complex():
           name = sum([[n, n] for n in name], [])


        for k, n in enumerate(name):

           prc = prcs[n][0]
           if prc == "None": continue
           name = "".join([tmp for tmp in prc if not tmp.isdigit()])
           A0 = get_block(A, k, k)
           cls = SparseSmootherCls[name][0]
           arg = SparseSmootherCls[name][1]
           if name == 'MUMPS':
               invA0 = cls(A0, gui=self.gui[prc], engine=self.engine)

           elif name.startswith('schur'):
               args = name.split("(")[-1].split(")")[0].split(",")
               dprint1("setting up schur for ", args)
               if len(args) > 1:
                   assert False, "not yet supported"
               for arg in args:
                    r1 = self.engine.dep_var_offset(arg.strip())
                    c1 = self.engine.r_dep_var_offset(arg.strip())
                    B  =  get_block(A, k, c1)
                    Bt =  get_block(A, r1, k)
                    B0 = get_block(A, r1, c1)
                    Md = mfem.Vector(M0.Height())
                    B0.GetDiag(Md)
                    for i in range(Md.Size()):
                        if Md[i] != 0.:
                            Bt.ScaleRow(i, 1/Md[i])
                        else:
                            assert False, "diagnal element of matrix is zero"

                    S = mfem.Mult(B, Bt)
                    invA0 = mfem.DSmoother(S)
                    invA0.iterative_mode = False

           else:
               invA0 = cls(A0, arg)
           invA0.iterative_mode = False
           M.SetDiagonalBlock(k, invA0)
        '''
        '''
        We should support Shur complement type preconditioner
        if offset.Size() > 2:
            B =  get_block(A, 1, 0)
            MinvBt = get_block(A, 0, 1)
            Md = mfem.Vector(get_block(A, 0, 0).Height())
            get_block(A, 0, 0).GetDiag(Md)
            for i in range(Md.Size()):
                if Md[i] != 0.:
                    MinvBt.ScaleRow(i, 1/Md[i])
                else:
                    assert False, "diagnal element of matrix is zero"
            S = mfem.Mult(B, MinvBt)
            S.iterative_mode = False
            SS = mfem.DSmoother(S)
            SS.iterative_mode = False
            M.SetDiagonalBlock(1, SS)
        '''

        '''
        int GMRES(const Operator &A, Vector &x, const Vector &b, Solver &M,
          int &max_iter, int m, double &tol, double atol, int printit)
        '''



'''
   eval_shape:

   this routine demonstrates how to evaluate shape function 
   with finer resolution

   (example)
    dof, ptx, shape_l2 = petram.helper.eval_shape.eval_shape(order=1, refine=10, fec='RT')
    petram.helper.eval_shape.plot_shape(dof, ptx, shape_l2)


'''
import numpy as np

from petram.mfem_config import use_parallel
if use_parallel:
   from petram.helper.mpi_recipes import *
   import mfem.par as mfem   
else:
   import mfem.ser as mfem

def eval_shape(order = 1, refine = 5, elem_type = 0,
               fec = 'ND'):

    if fec == 'ND':
        fec_type = mfem.ND_FECollection
        if order < 1: assert False, "ND order is 1 and above"        
    elif fec == 'RT':
        fec_type = mfem.RT_FECollection
    elif fec == 'H1':
        fec_type = mfem.H1_FECollection
        if order < 1: assert False, "H1 order is 1 and above"
    elif fec == 'L2':
        fec_type = mfem.L2_FECollection
    else:
        assert False, "unknown basis"
        
    if elem_type == 0:
        Nvert = 3; Nelem = 1; spaceDim = 2
    elif elem_type == 1:    
        Nvert = 4; Nelem = 1; spaceDim = 2

    mesh = mfem.Mesh(2, Nvert, Nelem, 0, spaceDim)

    if elem_type == 0:
        tri_v = [[1.,  0.3 ], [0.,  1.,], [0, 0]]
        tri_e = [[0, 1, 2], ]
        for j in range(Nvert):
            mesh.AddVertex(tri_v[j])
        for j in range(Nelem):
            mesh.AddTriangle(tri_e[j], j+1)
        mesh.FinalizeTriMesh(1,1, True)
    else:
        quad_v = [[-1, -1.3, ], [+1, -1, ], [+1, +1, ], [-1, +1,]]
        quad_e = [[0, 1, 2, 3]]
        for j in range(Nvert):
            mesh.AddVertex(quad_v[j])
        for j in range(Nelem):
            mesh.AddQuad(quad_e[j], j+1)
        mesh.FinalizeQuadMesh(1,1, True)

    #mesh.PrintToFile('plot_basis.mesh', 8)

    fe_coll = fec_type(order, spaceDim)
    fespace = mfem.FiniteElementSpace(mesh, fe_coll)

    x = mfem.GridFunction(fespace)
    x.Assign(0.0)
    x[0] = 1.0

    idx = 0
    geom = mesh.GetElementBaseGeometry(idx)
    T = mesh.GetElementTransformation(idx)
    fe = fespace.GetFE(idx)
    fe_nd = fe.GetDof()

    ir = fe.GetNodes()
    npt = ir.GetNPoints()
    dof = np.vstack([T.Transform(ir.IntPoint(i)) for i in range(npt)])
    
    RefG = mfem.GlobGeometryRefiner.Refine(geom, refine, 1);
    ir = RefG.RefPts

    npt = ir.GetNPoints()
    ptx = np.vstack([T.Transform(ir.IntPoint(i)) for i in range(npt)])

    shape = []
    if fec == 'ND' or fec == 'RT':
        mat = mfem.DenseMatrix(fe_nd, spaceDim)
        shape_func = fe.CalcVShape
        for i in range(npt):
            ip = ir.IntPoint(i)
            T.SetIntPoint(ip)
            fe.CalcVShape(T, mat)
            shape.append(mat.GetDataArray().copy())
    else:
        vec = mfem.Vector(fe_nd)
        for i in range(npt):
            ip = ir.IntPoint(i)            
            fe.CalcShape(ip, vec)
            shape.append(vec.GetDataArray().copy())

    return dof, ptx, np.stack(shape)

def plot_shape(dof, ptx,  shape, viewer = None):
    try:
        viewer.Validate()
    except:
        from ifigure.interactive import figure
        viewer = figure()

    ll = shape.shape[1]
    n = int(np.ceil(np.sqrt(ll)))
    
    viewer.cls()
    viewer.nsec(n, int(np.ceil(float(ll)/float(n))))

    for i in range(ll):
        x = ptx[:, 0]; y= ptx[:, 1]
        viewer.isec(i)
        if len(shape.shape) == 3:
            viewer.tripcolor(ptx[:,0], ptx[:,1], ptx[:,1]*0)            
            u=shape[:, i, 0]; v = shape[:, i, 1]            
            viewer.quiver(x, y, u, v)
        else:
            viewer.tripcolor(ptx[:,0], ptx[:,1], shape[:,i])
            idx = np.where(np.abs(shape[:,i]) < 1e-10)[0]
            if len(idx)>0: viewer.plot(ptx[idx,0], ptx[idx,1], 'gs')
        viewer.plot([dof[i][0]], [dof[i][1]], 'or')
    
if __name__ == '__main__':
    ans = eval_shape(fec = 'RT')
    for k in ans:
        print(k.shape)
